# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Callable, Optional, Sequence
import logging
import asyncio
import dataclasses
import pyuavcan.transport
from ._base import RedundantSession, RedundantSessionStatistics


_logger = logging.getLogger(__name__)


class RedundantFeedback(pyuavcan.transport.Feedback):
    """
    This is the output feedback extended with the reference to the inferior transport session
    that this feedback originates from.

    A redundant output session provides one feedback entry per inferior session;
    for example, if there are three inferiors in a redundant transport group,
    each outgoing transfer will generate three feedback entries
    (unless inferior sessions fail to provide their feedback entries for whatever reason).
    """

    def __init__(
        self, inferior_feedback: pyuavcan.transport.Feedback, inferior_session: pyuavcan.transport.OutputSession
    ):
        self._inferior_feedback = inferior_feedback
        self._inferior_session = inferior_session

    @property
    def original_transfer_timestamp(self) -> pyuavcan.transport.Timestamp:
        return self._inferior_feedback.original_transfer_timestamp

    @property
    def first_frame_transmission_timestamp(self) -> pyuavcan.transport.Timestamp:
        return self._inferior_feedback.first_frame_transmission_timestamp

    @property
    def inferior_feedback(self) -> pyuavcan.transport.Feedback:
        """
        The original feedback instance from the inferior session.
        """
        assert isinstance(self._inferior_feedback, pyuavcan.transport.Feedback)
        return self._inferior_feedback

    @property
    def inferior_session(self) -> pyuavcan.transport.OutputSession:
        """
        The inferior session that generated this feedback entry.
        """
        assert isinstance(self._inferior_session, pyuavcan.transport.OutputSession)
        return self._inferior_session


@dataclasses.dataclass(frozen=True)
class _WorkItem:
    """
    Send the transfer before the deadline, then notify the future unless it is already canceled.
    """

    transfer: pyuavcan.transport.Transfer
    monotonic_deadline: float
    future: asyncio.Future[bool]


@dataclasses.dataclass(frozen=True)
class _Inferior:
    """
    Each inferior runs a dedicated worker task.
    The worker takes work items from the queue one by one and attempts to transmit them.
    Upon completion (timeout/exception/success) the future is materialized unless cancelled.
    """

    session: pyuavcan.transport.OutputSession
    worker: asyncio.Task[None]
    queue: asyncio.Queue[_WorkItem]

    def close(self) -> None:
        try:
            self.session.close()
        finally:
            if self.worker.done():
                self.worker.result()
            else:
                self.worker.cancel()
            while True:
                try:
                    self.queue.get_nowait().future.cancel()
                except asyncio.QueueEmpty:
                    break


class RedundantOutputSession(RedundantSession, pyuavcan.transport.OutputSession):
    """
    This is a composite of a group of :class:`pyuavcan.transport.OutputSession`.
    Every outgoing transfer is simply forked into each of the inferior sessions.
    The result aggregation policy is documented in :func:`send`.
    """

    def __init__(
        self,
        specifier: pyuavcan.transport.OutputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        finalizer: Callable[[], None],
    ):
        """
        Do not call this directly! Use the factory method instead.
        """
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._finalizer: Optional[Callable[[], None]] = finalizer
        assert isinstance(self._specifier, pyuavcan.transport.OutputSessionSpecifier)
        assert isinstance(self._payload_metadata, pyuavcan.transport.PayloadMetadata)
        assert callable(self._finalizer)

        self._inferiors: list[_Inferior] = []
        self._feedback_handler: Optional[Callable[[RedundantFeedback], None]] = None
        self._idle_send_future: Optional[asyncio.Future[None]] = None
        self._lock = asyncio.Lock()

        self._stat_transfers = 0
        self._stat_payload_bytes = 0
        self._stat_errors = 0
        self._stat_drops = 0

    def _add_inferior(self, session: pyuavcan.transport.Session) -> None:
        assert isinstance(session, pyuavcan.transport.OutputSession)
        assert self._finalizer is not None, "The session was supposed to be unregistered"
        assert session.specifier == self.specifier and session.payload_metadata == self.payload_metadata
        if session in self.inferiors:
            return
        # Synchronize the feedback state.
        if self._feedback_handler is not None:
            self._enable_feedback_on_inferior(session)
        else:
            session.disable_feedback()
        # If all went well, add the new inferior to the set.
        que: asyncio.Queue[_WorkItem] = asyncio.Queue()
        tsk = asyncio.get_event_loop().create_task(self._inferior_worker_task(session, que))
        self._inferiors.append(_Inferior(session, tsk, que))
        # Unlock the pending transmission because now we have an inferior to work with.
        if self._idle_send_future is not None:
            self._idle_send_future.set_result(None)

    def _close_inferior(self, session_index: int) -> None:
        assert session_index >= 0, "Negative indexes may lead to unexpected side effects"
        assert self._finalizer is not None, "The session was supposed to be unregistered"
        try:
            session = self._inferiors.pop(session_index)
        except LookupError:
            pass
        else:
            session.close()  # May raise.

    @property
    def inferiors(self) -> Sequence[pyuavcan.transport.OutputSession]:
        return [x.session for x in self._inferiors]

    def enable_feedback(self, handler: Callable[[RedundantFeedback], None]) -> None:
        """
        The operation is atomic on all inferiors.
        If at least one inferior fails to enable feedback, all inferiors are rolled back into the disabled state.
        """
        self.disable_feedback()  # For state determinism.
        try:
            self._feedback_handler = handler
            for ses in self._inferiors:
                self._enable_feedback_on_inferior(ses.session)
        except Exception as ex:
            _logger.info("%s could not enable feedback, rolling back into the disabled state: %r", self, ex)
            self.disable_feedback()
            raise

    def disable_feedback(self) -> None:
        """
        The method implements the best-effort policy if any of the inferior sessions fail to disable feedback.
        """
        self._feedback_handler = None
        for ses in self._inferiors:
            try:
                ses.session.disable_feedback()
            except Exception as ex:
                _logger.exception("%s could not disable feedback on %r: %s", self, ses, ex)

    async def send(self, transfer: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
        """
        Sends the transfer via all of the inferior sessions concurrently.
        Returns when the first of the inferior calls succeeds; the remaining will keep sending in the background;
        that is, the redundant transport operates at the rate of the fastest inferior, delegating the slower ones
        to background tasks.
        Edge cases:

        - If there are no inferiors, the method will await until either the deadline is expired
          or an inferior(s) is (are) added. In the former case, the method returns False.
          In the latter case, the transfer is transmitted via the new inferior(s) using the remaining time
          until the deadline.

        - If at least one inferior succeeds, True is returned (logical OR).
          If the other inferiors raise exceptions, they are logged as errors and suppressed.

        - If all inferiors raise exceptions, one of them is propagated, the rest are logged as errors and suppressed.

        - If all inferiors time out, False is returned (logical OR).

        In other words, the error handling strategy is optimistic: if one inferior reported success,
        the call is assumed to have succeeded; best result is always returned.
        """
        if self._finalizer is None:
            raise pyuavcan.transport.ResourceClosedError(f"{self} is closed")

        loop = asyncio.get_running_loop()
        async with self._lock:  # Serialize access to the inferiors and the idle future.
            # It is required to create a local copy to prevent disruption of the logic when
            # the set of inferiors is changed in the background. Oh, Rust, where art thou.
            inferiors = list(self._inferiors)

            # This part is a bit tricky. If there are no inferiors, we have nowhere to send the transfer.
            # Instead of returning immediately, we hang out here until the deadline is expired hoping that
            # an inferior is added while we're waiting here.
            assert not self._idle_send_future
            if not inferiors and monotonic_deadline > loop.time():
                try:
                    _logger.debug("%s has no inferiors; suspending the send method...", self)
                    self._idle_send_future = loop.create_future()
                    try:
                        await asyncio.wait_for(self._idle_send_future, timeout=monotonic_deadline - loop.time())
                    except asyncio.TimeoutError:
                        pass
                    else:
                        self._idle_send_future.result()  # Collect the empty result to prevent asyncio from complaining.
                    # The set of inferiors may have been updated.
                    inferiors = list(self._inferiors)
                    _logger.debug(
                        "%s send method unsuspended; available inferiors: %r; remaining time: %f",
                        self,
                        inferiors,
                        monotonic_deadline - loop.time(),
                    )
                finally:
                    self._idle_send_future = None
            assert not self._idle_send_future
            if not inferiors:
                self._stat_drops += 1
                return False  # Still nothing.

            # We have at least one inferior so we can handle this transaction. Create the work items.
            futures: list[asyncio.Future[bool]] = []
            for inf in self._inferiors:
                fut: asyncio.Future[bool] = asyncio.Future()
                inf.queue.put_nowait(_WorkItem(transfer, monotonic_deadline, fut))
                futures.append(fut)

            # Execute the work items concurrently and unblock as soon as the first inferior is done transmitting.
            # Those that are still pending are cancelled because we're not going to wait around for the slow ones.
            done, pending = await asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED)
            assert len(done) + len(pending) == len(inferiors) == len(futures)
            _logger.debug("%s send results: done=%s, pending=%s", self, done, pending)
            for p in pending:
                p.cancel()  # We will no longer need this.

            # Extract the results to determine the final outcome of the transaction.
            results = [x.result() for x in done if x.exception() is None]
            exceptions = [x.exception() for x in done if x.exception() is not None]
            assert 0 < (len(results) + len(exceptions)) <= len(inferiors)  # Some tasks may be not yet done.
            assert not results or all(isinstance(x, bool) for x in results)
            if exceptions and not results:
                self._stat_errors += 1
                exc = exceptions[0]
                assert isinstance(exc, BaseException)
                raise exc
            if results and any(results):
                self._stat_transfers += 1
                self._stat_payload_bytes += sum(map(len, transfer.fragmented_payload))
                return True
            self._stat_drops += 1
            return False

    @property
    def specifier(self) -> pyuavcan.transport.OutputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> RedundantSessionStatistics:
        """
        - ``transfers``     - the number of redundant transfers where at least ONE inferior succeeded (success count).
        - ``errors``        - the number of redundant transfers where ALL inferiors raised exceptions (failure count).
        - ``payload_bytes`` - the number of payload bytes in successful redundant transfers counted in ``transfers``.
        - ``drops``         - the number of redundant transfers where ALL inferiors timed out (timeout count).
        - ``frames``        - the total number of frames summed from all inferiors (i.e., replicated frame count).
          This value is invalidated when the set of inferiors is changed. The semantics may change later.
        """
        inferiors = [s.session.sample_statistics() for s in self._inferiors]
        return RedundantSessionStatistics(
            transfers=self._stat_transfers,
            frames=sum(s.frames for s in inferiors),
            payload_bytes=self._stat_payload_bytes,
            errors=self._stat_errors,
            drops=self._stat_drops,
            inferiors=inferiors,
        )

    def close(self) -> None:
        for s in self._inferiors:
            try:
                s.close()
            except Exception as ex:
                _logger.exception("%s could not close inferior %s: %s", self, s, ex)
        self._inferiors.clear()

        fin, self._finalizer = self._finalizer, None
        if fin is not None:
            fin()

    async def _inferior_worker_task(self, ses: pyuavcan.transport.OutputSession, que: asyncio.Queue[_WorkItem]) -> None:
        try:
            _logger.debug("%s: Task for inferior %r is starting", self, ses)
            while self._finalizer:
                wrk = await que.get()
                try:
                    result = await ses.send(wrk.transfer, wrk.monotonic_deadline)
                except (asyncio.CancelledError, pyuavcan.transport.ResourceClosedError):
                    break  # Do not cancel the future because we don't want to unblock the master task.
                except Exception as ex:
                    _logger.error("%s: Inferior %r failed: %s: %s", self, ses, type(ex).__name__, ex)
                    _logger.debug("%s: Stack trace for the above inferior failure:", self, exc_info=True)
                    if not wrk.future.done():
                        wrk.future.set_exception(ex)
                else:
                    _logger.debug(
                        "%s: Inferior %r send result: %s; future %s",
                        self,
                        ses,
                        "success" if result else "timeout",
                        wrk.future,
                    )
                    if not wrk.future.done():
                        wrk.future.set_result(result)
        except (asyncio.CancelledError, pyuavcan.transport.ResourceClosedError):
            pass
        except Exception as ex:
            _logger.exception("%s: Task for %r has encountered an unhandled exception: %s", self, ses, ex)
        finally:
            _logger.debug("%s: Task for %r is stopping", self, ses)

    def _enable_feedback_on_inferior(self, inferior_session: pyuavcan.transport.OutputSession) -> None:
        def proxy(fb: pyuavcan.transport.Feedback) -> None:
            """
            Intercepts a feedback report from an inferior session,
            constructs a higher-level redundant feedback instance from it,
            and then passes it along to the higher-level handler.
            """
            if inferior_session not in self.inferiors:
                _logger.warning(
                    "%s got unexpected feedback %s from %s which is not a registered inferior. "
                    "The transport or its underlying software or hardware are probably misbehaving, "
                    "or this inferior has just been removed.",
                    self,
                    fb,
                    inferior_session,
                )
                return

            handler = self._feedback_handler
            if handler is not None:
                new_fb = RedundantFeedback(fb, inferior_session)
                try:
                    handler(new_fb)
                except Exception as ex:
                    _logger.exception("%s: Unhandled exception in the feedback handler %s: %s", self, handler, ex)
            else:
                _logger.debug("%s ignoring unattended feedback %r from %r", self, fb, inferior_session)

        inferior_session.enable_feedback(proxy)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import itertools
import dataclasses
from ._frame import FrameFormat


@dataclasses.dataclass(frozen=True)
class FilterConfiguration:
    identifier: int
    """The reference CAN ID value."""

    mask: int
    """Mask applies to the identifier only. It does not contain any special flags."""

    format: typing.Optional[FrameFormat]
    """None means no preference -- both formats will be accepted."""

    def __post_init__(self) -> None:
        max_bit_length = 2 ** self.identifier_bit_length - 1
        if not (0 <= self.identifier <= max_bit_length):
            raise ValueError(f"Invalid identifier: {self.identifier}")
        if not (0 <= self.mask <= max_bit_length):
            raise ValueError(f"Invalid mask: {self.mask}")

    @property
    def identifier_bit_length(self) -> int:
        # noinspection PyTypeChecker
        return int(self.format if self.format is not None else max(FrameFormat))

    @staticmethod
    def new_promiscuous(frame_format: typing.Optional[FrameFormat] = None) -> FilterConfiguration:
        """
        Returns a configuration that accepts all frames of the specified format.
        If the format is not specified, no distinction will be made.
        Note that some CAN controllers may have difficulty supporting both formats on a single filter.
        """
        return FilterConfiguration(identifier=0, mask=0, format=frame_format)

    @property
    def rank(self) -> int:
        """
        This is the number of set bits in the mask.
        This is a part of the CAN acceptance filter configuration optimization algorithm;
        see :func:`optimize_filter_configurations`.

        We return negative rank for configurations which do not distinguish between extended and base frames
        in order to discourage merger of configurations of different frame types, since they are hard to
        support in certain CAN controllers. The effect of this is that we guarantee that an ambivalent filter
        configuration will never appear if the controller has at least two acceptance filters.
        Negative rank is computed by subtracting the number of bits in the CAN ID
        (or 29 if the filter accepts both base and extended identifiers) from the original rank.
        """
        mask_mask = 2 ** self.identifier_bit_length - 1
        rank = bin(self.mask & mask_mask).count("1")
        if self.format is None:
            rank -= int(self.identifier_bit_length)  # Discourage merger of ambivalent filters.
        return rank

    def merge(self, other: FilterConfiguration) -> FilterConfiguration:
        """
        This is a part of the CAN acceptance filter configuration optimization algorithm;
        see :func:`optimize_filter_configurations`.

        Given two filter configurations ``A`` and ``B``, where ``A`` accepts CAN frames whose identifiers
        belong to ``Ca`` and likewise ``Cb`` for ``B``, the merge product of ``A`` and ``B`` would be a
        new filter configuration that accepts CAN frames belonging to a new set which is a superset of
        the union of ``Ca`` and ``Cb``.
        """
        mask = self.mask & other.mask & ~(self.identifier ^ other.identifier)
        identifier = self.identifier & mask
        fmt = self.format if self.format == other.format else None
        return FilterConfiguration(identifier=identifier, mask=mask, format=fmt)

    def __str__(self) -> str:
        out = "".join(
            (str((self.identifier >> bit) & 1) if self.mask & (1 << bit) != 0 else "x")
            for bit in reversed(range(int(self.format or FrameFormat.EXTENDED)))
        )
        return (self.format.name[:3].lower() if self.format else "any") + ":" + out


def optimize_filter_configurations(
    configurations: typing.Iterable[FilterConfiguration], target_number_of_configurations: int
) -> typing.Sequence[FilterConfiguration]:
    """
    Implements the CAN acceptance filter configuration optimization algorithm described in the Specification.
    The algorithm was originally proposed by P. Kirienko and I. Sheremet.

    Given a
    set of ``K``  filter configurations that accept CAN frames whose identifiers belong to the set ``C``,
    and ``N`` acceptance filters implemented in hardware, where ``1 <= N < K``, find a new
    set of ``K'`` filter configurations that accept CAN frames whose identifiers belong to the set ``C'``,
    such that ``K' <= N``, ``C'`` is a superset of ``C``, and ``|C'|`` is minimized.

    The algorithm is not defined for ``N >= K`` because this configuration is considered optimal.
    The function returns the input set unchanged in this case.
    If the target number of configurations is not positive, a ValueError is raised.

    The time complexity of this implementation is ``O(K!)``; it should be optimized.
    """
    if target_number_of_configurations < 1:
        raise ValueError(f"The number of configurations must be positive; found {target_number_of_configurations}")

    configurations = list(configurations)
    while len(configurations) > target_number_of_configurations:
        options = itertools.starmap(
            lambda ia, ib: (ia[0], ib[0], ia[1].merge(ib[1])), itertools.permutations(enumerate(configurations), 2)
        )
        index_replace, index_remove, merged = max(options, key=lambda x: int(x[2].rank))
        configurations[index_replace] = merged
        del configurations[index_remove]  # Invalidates indexes

    assert all(map(lambda x: isinstance(x, FilterConfiguration), configurations))
    return configurations


def _unittest_can_media_filter_faults() -> None:
    from pytest import raises

    with raises(ValueError):
        FilterConfiguration(0, -1, None)

    with raises(ValueError):
        FilterConfiguration(-1, 0, None)

    for fmt in FrameFormat:
        with raises(ValueError):
            FilterConfiguration(2 ** int(fmt), 0, fmt)

        with raises(ValueError):
            FilterConfiguration(0, 2 ** int(fmt), fmt)

    with raises(ValueError):
        optimize_filter_configurations([], 0)


# noinspection SpellCheckingInspection
def _unittest_can_media_filter_str() -> None:
    assert str(FilterConfiguration(0b10101010, 0b11101000, FrameFormat.EXTENDED)) == "ext:xxxxxxxxxxxxxxxxxxxxx101x1xxx"

    assert (
        str(FilterConfiguration(0b10101010101010101010101010101, 0b10111111111111111111111111111, FrameFormat.EXTENDED))
        == "ext:1x101010101010101010101010101"
    )

    assert str(FilterConfiguration(0b10101010101, 0b11111111111, FrameFormat.BASE)) == "bas:10101010101"

    assert str(FilterConfiguration(123, 456, None)) == "any:xxxxxxxxxxxxxxxxxxxx001xx1xxx"

    assert str(FilterConfiguration.new_promiscuous()) == "any:xxxxxxxxxxxxxxxxxxxxxxxxxxxxx"

    assert repr(FilterConfiguration(123, 456, None)) == "FilterConfiguration(identifier=123, mask=456, format=None)"


def _unittest_can_media_filter_merge() -> None:
    assert FilterConfiguration(123456, 0, None).rank == -29  # Worst rank
    assert FilterConfiguration(123456, 0b110, None).rank == -27  # Two better

    assert FilterConfiguration(1234, 0b110, FrameFormat.BASE).rank == 2

    assert (
        FilterConfiguration(0b111, 0b111, FrameFormat.EXTENDED)
        .merge(FilterConfiguration(0b111, 0b111, FrameFormat.BASE))
        .rank
        == -29 + 3
    )



# Copyright (C) 2021  UAVCAN Consortium  <uavcan.org>
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Optional
import logging
import pyuavcan
import pyuavcan.application
from pyuavcan.presentation import ServiceRequestMetadata
from uavcan.register import Access_1 as Access
from uavcan.register import List_1 as List
from uavcan.register import Name_1 as Name
from .register import ValueConversionError, ValueProxyWithFlags


class RegisterServer:
    # noinspection PyUnresolvedReferences,PyTypeChecker
    """
    Implementation of the standard network service ``uavcan.register``; specifically, List and Access.

    This server implements automatic type conversion by invoking
    :meth:`pyuavcan.application.register.ValueProxy.assign` on every set request.
    This means that, for example, one can successfully modify a register of type
    ``bool[x]`` by sending a set request of type ``real64[x]``, or ``string`` with ``unstructured``, etc.

    Here is a demo. Set up a node -- it will instantiate a register server automatically:

    ..  doctest::
        :hide:

        >>> import tests
        >>> tests.asyncio_allow_event_loop_access_from_top_level()
        >>> from tests import doctest_await

    >>> import pyuavcan
    >>> from pyuavcan.transport.loopback import LoopbackTransport
    >>> from pyuavcan.application.register import Registry, Value, ValueProxy, Integer64, Real16, Unstructured
    >>> node = pyuavcan.application.make_node(pyuavcan.application.NodeInfo(), transport=LoopbackTransport(1))
    >>> node.registry.setdefault("foo", Value(integer64=Integer64([1, 20, -100]))).ints
    [1, 20, -100]
    >>> node.start()

    List registers:

    >>> import uavcan.register
    >>> cln_list = node.make_client(uavcan.register.List_1, server_node_id=1)
    >>> response, _ = doctest_await(cln_list.call(uavcan.register.List_1.Request(index=0)))
    >>> response.name.name.tobytes().decode()   # The dummy register we created above.
    'foo'
    >>> response, _ = doctest_await(cln_list.call(uavcan.register.List_1.Request(index=99)))
    >>> response.name.name.tobytes().decode()   # Out of range -- empty string returned to indicate that.
    ''

    Get the dummy register created above:

    >>> cln_access = node.make_client(uavcan.register.Access_1, server_node_id=1)
    >>> request = uavcan.register.Access_1.Request()
    >>> request.name.name = "foo"
    >>> response, _ = doctest_await(cln_access.call(request))
    >>> response.mutable, response.persistent
    (True, False)
    >>> ValueProxy(response.value).ints
    [1, 20, -100]

    Set a new value and read it back.
    Notice that the type does not match but it is automatically converted by the server.

    >>> request.value.real16 = Real16([3.14159, 2.71828, -500])  # <-- the type is different but it's okay.
    >>> response, _ = doctest_await(cln_access.call(request))
    >>> ValueProxy(response.value).ints     # Automatically converted.
    [3, 3, -500]
    >>> node.registry["foo"].ints           # Yup, the register is, indeed, updated by the server.
    [3, 3, -500]

    If the type cannot be converted or the register is immutable, the write is ignored,
    as prescribed by the register network service definition:

    >>> request.value.unstructured = Unstructured(b'Hello world!')
    >>> response, _ = doctest_await(cln_access.call(request))
    >>> ValueProxy(response.value).ints  # Conversion is not possible, same value retained.
    [3, 3, -500]

    An attempt to access a non-existent register returns an empty value:

    >>> request.name.name = 'bar'
    >>> response, _ = doctest_await(cln_access.call(request))
    >>> response.value.empty is not None
    True

    >>> node.close()
    """

    def __init__(self, node: pyuavcan.application.Node) -> None:
        """
        :param node: The node instance to serve the register API for.
        """
        self._node = node

        srv_list = self.node.get_server(List)
        srv_access = self.node.get_server(Access)

        def start() -> None:
            srv_list.serve_in_background(self._handle_list)
            srv_access.serve_in_background(self._handle_access)

        def close() -> None:
            srv_list.close()
            srv_access.close()

        node.add_lifetime_hooks(start, close)

    @property
    def node(self) -> pyuavcan.application.Node:
        return self._node

    async def _handle_list(self, request: List.Request, metadata: ServiceRequestMetadata) -> List.Response:
        name = self.node.registry.index(request.index)
        _logger.debug("%r: List request index %r name %r %r", self, request.index, name, metadata)
        if name is not None:
            return List.Response(Name(name))
        return List.Response()

    async def _handle_access(self, request: Access.Request, metadata: ServiceRequestMetadata) -> Access.Response:
        name = request.name.name.tobytes().decode("utf8", "ignore")
        try:
            v: Optional[ValueProxyWithFlags] = self.node.registry[name]
        except KeyError:
            v = None

        if v is not None and v.mutable and not request.value.empty:
            try:
                v.assign(request.value)
                self.node.registry[name] = v
            except ValueConversionError as ex:
                _logger.debug("%r: Conversion from %r to %r is not possible: %s", self, request.value, v.value, ex)
            # Read back one more time just in case to confirm write.
            try:
                v = self.node.registry[name]
            except KeyError:
                v = None

        if v is not None:
            response = Access.Response(
                mutable=v.mutable,
                persistent=v.persistent,
                value=v.value,
            )
        else:
            response = Access.Response()  # No such register
        _logger.debug("%r: Access %r: %r %r", self, metadata, request, response)
        return response


_logger = logging.getLogger(__name__)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import numpy

Byte = numpy.uint8
"""
We must use uint8 instead of ubyte because uint8 is platform-invariant whereas (u)byte is platform-dependent.
"""

StdPrimitive = typing.TypeVar(
    "StdPrimitive",
    numpy.float64,
    numpy.float32,
    numpy.float16,
    numpy.uint8,
    numpy.uint16,
    numpy.uint32,
    numpy.uint64,
    numpy.int8,
    numpy.int16,
    numpy.int32,
    numpy.int64,
)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import enum
import dataclasses


@dataclasses.dataclass(frozen=True)
class DataSpecifier:
    """
    The data specifier defines what category and type of data is exchanged over a transport session.
    See the abstract transport model for details.
    """


@dataclasses.dataclass(frozen=True)
class MessageDataSpecifier(DataSpecifier):
    SUBJECT_ID_MASK = 2 ** 13 - 1

    subject_id: int

    def __post_init__(self) -> None:
        if not (0 <= self.subject_id <= self.SUBJECT_ID_MASK):
            raise ValueError(f"Invalid subject ID: {self.subject_id}")


@dataclasses.dataclass(frozen=True)
class ServiceDataSpecifier(DataSpecifier):
    class Role(enum.Enum):
        REQUEST = enum.auto()
        """
        Request output role is for clients.
        Request input role is for servers.
        """
        RESPONSE = enum.auto()
        """
        Response output role is for servers.
        Response input role is for clients.
        """

    SERVICE_ID_MASK = 2 ** 9 - 1

    service_id: int
    role: Role

    def __post_init__(self) -> None:
        assert self.role in self.Role
        if not (0 <= self.service_id <= self.SERVICE_ID_MASK):
            raise ValueError(f"Invalid service ID: {self.service_id}")



# Copyright (C) 2021  UAVCAN Consortium  <uavcan.org>
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Union, Iterable, List, Any, Optional, no_type_check
from numpy.typing import NDArray
import pyuavcan
from pyuavcan.dsdl import get_attribute
from .backend import Value as Value
from . import String, Unstructured, Bit
from . import Integer8, Integer16, Integer32, Integer64
from . import Natural8, Natural16, Natural32, Natural64
from . import Real16, Real32, Real64


class ValueConversionError(ValueError):
    """
    Raised when there is no known conversion between the argument and the specified register.
    """


VALUE_OPTION_NAMES = [x for x in dir(Value) if not x.startswith("_")]


class ValueProxy:
    """
    This a wrapper over the standard ``uavcan.register.Value`` (transpiled into :class:`Value`)
    with convenience accessors added that enable automatic conversion (with implicit casting)
    between native Python types and DSDL types.

    It is possible to create a new instance from native types,
    in which case the most suitable regiter type will be deduced automatically.
    Do not rely on this behavior if a specific register type needs to be ensured.

    >>> from pyuavcan.application.register import Real64, Bit, String, Unstructured
    >>> p = ValueProxy(Value(bit=Bit([True, False])))   # Specify explicit type.
    >>> p.bools
    [True, False]
    >>> p.ints
    [1, 0]
    >>> p.floats
    [1.0, 0.0]
    >>> p.assign([0, 1.0])
    >>> p.bools
    [False, True]

    >>> p = ValueProxy([0, 1.5, 2.3, -9])               # Use deduction.
    >>> p.floats
    [0.0, 1.5, 2.3, -9.0]
    >>> p.ints
    [0, 2, 2, -9]
    >>> p.bools
    [False, True, True, True]
    >>> p.assign([False, True, False, True])
    >>> p.floats
    [0.0, 1.0, 0.0, 1.0]

    >>> p = ValueProxy(False)
    >>> bool(p)
    False
    >>> int(p)
    0
    >>> float(p)
    0.0
    >>> p.assign(1)
    >>> bool(p), int(p), float(p)
    (True, 1, 1.0)

    >>> p = ValueProxy("Hello world!")                  # Create string-typed register value.
    >>> str(p)
    'Hello world!'
    >>> bytes(p)
    b'Hello world!'
    >>> p.assign('Another string')
    >>> str(p)
    'Another string'
    >>> bytes(p)
    b'Another string'

    >>> p = ValueProxy(b"ab01")                         # Create unstructured-typed register value.
    >>> str(p)
    'ab01'
    >>> bytes(p)
    b'ab01'
    >>> p.assign("String implicitly converted to bytes")
    >>> bytes(p)
    b'String implicitly converted to bytes'
    """

    def __init__(self, v: RelaxedValue) -> None:
        """
        Accepts a wide set of native and generated types.
        Passing native values is not recommended because the type deduction logic may be changed in the future.
        To ensure stability, pass only values of ``uavcan.primitive.*``, or :class:`Value`, or :class:`ValueProxy`.

        >>> list(ValueProxy(Value(natural16=Natural16([123, 456]))).value.natural16.value)  # Explicit Value.
        [123, 456]
        >>> list(ValueProxy(Natural16([123, 456])).value.natural16.value)                   # Same as above.
        [123, 456]
        >>> ValueProxy(-123).value.integer64.value[0]               # Integers default to 64-bit.
        -123
        >>> list(ValueProxy([-1.23, False]).value.real64.value)     # Floats also default to 64-bit.
        [-1.23, 0.0]
        >>> list(ValueProxy([True, False]).value.bit.value)         # Booleans default to bits.
        [True, False]
        >>> ValueProxy(b"Hello unstructured!").value.unstructured.value.tobytes()   # Bytes to unstructured.
        b'Hello unstructured!'

        And so on...

        :raises: :class:`ValueConversionError` if the conversion is impossible or ambiguous.
        """
        from copy import copy

        self._value = copy(_strictify(v))

    @property
    def value(self) -> Value:
        """Access to the underlying standard DSDL type ``uavcan.register.Value``."""
        return self._value

    def assign(self, source: RelaxedValue) -> None:
        """
        Converts the value from the source into the type of the current instance, and updates this instance.

        :raises: :class:`ValueConversionError` if the source value cannot be converted to the register's type.
        """
        opt_to = _get_option_name(self._value)
        res = _do_convert(self._value, _strictify(source))
        if res is None:
            raise ValueConversionError(f"Source {source!r} cannot be assigned to {self!r}")
        assert _get_option_name(res) == opt_to
        self._value = res

    def assign_environment_variable(self, environment_variable_value: Union[str, bytes]) -> None:
        """
        This is like :meth:`assign`, but the argument is the value of an environment variable.
        The conversion rules are documented in the standard RPC-service specification ``uavcan.register.Access``.
        See also: :func:`pyuavcan.application.register.get_environment_variable_name`.

        :param environment_variable_value: E.g., ``1 2 3``.
        :raises: :class:`ValueConversionError` if the value cannot be converted.
        """
        if self.value.empty or self.value.string or self.value.unstructured:
            self.assign(environment_variable_value)
        else:
            numbers: List[Union[int, float]] = []
            for nt in environment_variable_value.split():
                try:
                    numbers.append(int(nt))
                except ValueError:
                    try:
                        numbers.append(float(nt))
                    except ValueError:
                        raise ValueConversionError(
                            f"Cannot update {self!r} from environment value {environment_variable_value!r}"
                        ) from None
            self.assign(numbers)

    @property
    def floats(self) -> List[float]:
        """
        Converts the value to a list of floats, or raises :class:`ValueConversionError` if not possible.
        """
        # pylint: disable=multiple-statements

        def cast(a: Any) -> List[float]:
            return [float(x) for x in a.value]

        v = self._value
        # fmt: off
        if v.bit:       return cast(v.bit)
        if v.integer8:  return cast(v.integer8)
        if v.integer16: return cast(v.integer16)
        if v.integer32: return cast(v.integer32)
        if v.integer64: return cast(v.integer64)
        if v.natural8:  return cast(v.natural8)
        if v.natural16: return cast(v.natural16)
        if v.natural32: return cast(v.natural32)
        if v.natural64: return cast(v.natural64)
        if v.real16:    return cast(v.real16)
        if v.real32:    return cast(v.real32)
        if v.real64:    return cast(v.real64)
        # fmt: on
        raise ValueConversionError(f"{v!r} cannot be represented numerically")

    @property
    def ints(self) -> List[int]:
        """
        Converts the value to a list of ints, or raises :class:`ValueConversionError` if not possible.
        """
        return [round(x) for x in self.floats]

    @property
    def bools(self) -> List[bool]:
        """
        Converts the value to a list of bools, or raises :class:`ValueConversionError` if not possible.
        """
        return [bool(x) for x in self.ints]

    def __float__(self) -> float:
        """Takes the first item from :attr:`floats`."""
        return self.floats[0]

    def __int__(self) -> int:
        """Takes the first item from :attr:`ints`."""
        return round(float(self))

    def __bool__(self) -> bool:
        """Takes the first item from :attr:`bools`."""
        return bool(int(self))

    def __str__(self) -> str:
        v = self._value
        if v.empty:
            return ""
        if v.string:
            return str(v.string.value.tobytes().decode("utf8"))
        if v.unstructured:
            return str(v.unstructured.value.tobytes().decode("utf8", "ignore"))
        raise ValueConversionError(f"{v!r} cannot be converted to string")

    def __bytes__(self) -> bytes:
        v = self._value
        if v.empty:
            return b""
        if v.string:
            return bytes(v.string.value.tobytes())
        if v.unstructured:
            return bytes(v.unstructured.value.tobytes())
        raise ValueConversionError(f"{v!r} cannot be converted to bytes")

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, repr(self._value))


RelaxedValue = Union[
    # Explicit values
    ValueProxy,
    Value,
    # Value option types
    String,
    Unstructured,
    Bit,
    Integer8,
    Integer16,
    Integer32,
    Integer64,
    Natural8,
    Natural16,
    Natural32,
    Natural64,
    Real16,
    Real32,
    Real64,
    # Native types
    str,
    bytes,
    bool,
    int,
    float,
    # Native collections
    Iterable[bool],
    Iterable[int],
    Iterable[float],
    NDArray[Any],
]
"""
These types can be automatically converted to :class:`Value` with a particular option selected.
"""


def _do_convert(to: Value, s: Value) -> Optional[Value]:
    """
    This is a bit rough around the edges; consider it to be an MVP.
    """
    # pylint: disable=multiple-statements
    if to.empty or s.empty:  # Everything is convertible to empty, and empty is convertible to everything.
        return to
    if (to.string and s.string) or (to.unstructured and s.unstructured):
        return s
    if to.string and s.unstructured:
        return Value(string=String(s.unstructured.value))
    if to.unstructured and s.string:
        return Value(unstructured=Unstructured(s.string.value))

    if s.string or s.unstructured:
        return None

    val_s: NDArray[Any] = get_attribute(
        s,
        _get_option_name(s),
    ).value.copy()
    val_s.resize(
        get_attribute(to, _get_option_name(to)).value.size,
        refcheck=False,
    )
    # At this point it is known that both values are of the same dimension.
    # fmt: off
    if to.bit:    return Value(bit=Bit([x != 0 for x in val_s]))
    if to.real16: return Value(real16=Real16(val_s))
    if to.real32: return Value(real32=Real32(val_s))
    if to.real64: return Value(real64=Real64(val_s))
    # fmt: on
    val_s_int = [round(x) for x in val_s]
    del val_s
    # fmt: off
    if to.integer8:  return Value(integer8=Integer8(val_s_int))
    if to.integer16: return Value(integer16=Integer16(val_s_int))
    if to.integer32: return Value(integer32=Integer32(val_s_int))
    if to.integer64: return Value(integer64=Integer64(val_s_int))
    if to.natural8:  return Value(natural8=Natural8(val_s_int))
    if to.natural16: return Value(natural16=Natural16(val_s_int))
    if to.natural32: return Value(natural32=Natural32(val_s_int))
    if to.natural64: return Value(natural64=Natural64(val_s_int))
    # fmt: on
    assert False


def _strictify(s: RelaxedValue) -> Value:
    # pylint: disable=multiple-statements,too-many-branches
    # fmt: off
    if isinstance(s, Value):                return s
    if isinstance(s, ValueProxy):           return s.value
    if isinstance(s, (bool, int, float)):   return _strictify([s])
    if isinstance(s, str):                  return _strictify(String(s))
    if isinstance(s, bytes):                return _strictify(Unstructured(s))
    # fmt: on
    # fmt: off
    if isinstance(s, String):       return Value(string=s)
    if isinstance(s, Unstructured): return Value(unstructured=s)
    if isinstance(s, Bit):          return Value(bit=s)
    if isinstance(s, Integer8):     return Value(integer8=s)
    if isinstance(s, Integer16):    return Value(integer16=s)
    if isinstance(s, Integer32):    return Value(integer32=s)
    if isinstance(s, Integer64):    return Value(integer64=s)
    if isinstance(s, Natural8):     return Value(natural8=s)
    if isinstance(s, Natural16):    return Value(natural16=s)
    if isinstance(s, Natural32):    return Value(natural32=s)
    if isinstance(s, Natural64):    return Value(natural64=s)
    if isinstance(s, Real16):       return Value(real16=s)
    if isinstance(s, Real32):       return Value(real32=s)
    if isinstance(s, Real64):       return Value(real64=s)
    # fmt: on

    s = list(s)
    if not s:
        return Value()  # Empty list generalized into Value.empty.
    if all(isinstance(x, bool) for x in s):
        return _strictify(Bit(s))  # type: ignore
    if all(isinstance(x, (int, bool)) for x in s):
        return _strictify(Natural64(s)) if all(x >= 0 for x in s) else _strictify(Integer64(s))  # type: ignore
    if all(isinstance(x, (float, int, bool)) for x in s):
        return _strictify(Real64(s))

    raise ValueConversionError(f"Don't know how to convert {s!r} into {Value}")  # pragma: no cover


def _get_option_name(x: Value) -> str:
    for n in VALUE_OPTION_NAMES:
        if get_attribute(x, n):
            return n
    raise TypeError(f"Invalid value: {x!r}; expected option names: {VALUE_OPTION_NAMES}")  # pragma: no cover


@no_type_check
def _unittest_strictify() -> None:
    import pytest

    v = Value(string=String("abc"))
    assert v is _strictify(v)  # Transparency.
    assert repr(v) == repr(_strictify(ValueProxy(v)))

    assert list(_strictify(+1).natural64.value) == [+1]
    assert list(_strictify(-1).integer64.value) == [-1]
    assert list(_strictify(1.1).real64.value) == [pytest.approx(1.1)]
    assert list(_strictify(True).bit.value) == [True]
    assert _strictify([]).empty

    assert _strictify("Hello").string.value.tobytes().decode() == "Hello"
    assert _strictify(b"Hello").unstructured.value.tobytes() == b"Hello"


@no_type_check
def _unittest_convert() -> None:
    import pytest

    q = Value

    def _once(a: Value, b: RelaxedValue) -> Value:
        c = ValueProxy(a)
        c.assign(b)
        return c.value

    assert _once(q(), q()).empty
    assert _once(q(), String("Hello")).empty
    assert _once(q(string=String("A")), String("B")).string.value.tobytes().decode() == "B"
    assert _once(q(string=String("A")), Unstructured(b"B")).string.value.tobytes().decode() == "B"
    assert list(_once(q(natural16=Natural16([1, 2])), Natural64([1, 2])).natural16.value) == [1, 2]

    assert list(_once(q(bit=Bit([False, False])), Integer32([-1, 0])).bit.value) == [True, False]
    assert list(_once(q(integer8=Integer8([0, 1])), Real64([3.3, 6.4])).integer8.value) == [3, 6]
    assert list(_once(q(integer16=Integer16([0, 1])), Real32([3.3, 6.4])).integer16.value) == [3, 6]
    assert list(_once(q(integer32=Integer32([0, 1])), Real16([3.3, 6.4])).integer32.value) == [3, 6]
    assert list(_once(q(integer64=Integer64([0, 1])), Real64([3.3, 6.4])).integer64.value) == [3, 6]
    assert list(_once(q(natural8=Natural8([0, 1])), Real64([3.3, 6.4])).natural8.value) == [3, 6]
    assert list(_once(q(natural16=Natural16([0, 1])), Real64([3.3, 6.4])).natural16.value) == [3, 6]
    assert list(_once(q(natural32=Natural32([0, 1])), Real64([3.3, 6.4])).natural32.value) == [3, 6]
    assert list(_once(q(natural64=Natural64([0, 1])), Real64([3.3, 6.4])).natural64.value) == [3, 6]
    assert list(_once(q(real16=Real16([0])), Bit([True])).real16.value) == [pytest.approx(1.0)]
    assert list(_once(q(real32=Real32([0])), Bit([True])).real32.value) == [pytest.approx(1.0)]
    assert list(_once(q(real64=Real64([0])), Bit([True])).real64.value) == [pytest.approx(1.0)]



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import copy
import typing
import logging
import dataclasses
import pyuavcan.transport
from pyuavcan.transport import Timestamp
from .._frame import UAVCANFrame, TRANSFER_ID_MODULO
from .._identifier import CANID, MessageCANID, ServiceCANID
from ._base import CANSession, SessionFinalizer
from ._transfer_sender import serialize_transfer


@dataclasses.dataclass(frozen=True)
class SendTransaction:
    frames: typing.List[UAVCANFrame]
    loopback_first: bool
    monotonic_deadline: float


SendHandler = typing.Callable[[SendTransaction], typing.Awaitable[bool]]

_logger = logging.getLogger(__name__)


class CANFeedback(pyuavcan.transport.Feedback):
    def __init__(self, original_transfer_timestamp: Timestamp, first_frame_transmission_timestamp: Timestamp):
        self._original_transfer_timestamp = original_transfer_timestamp
        self._first_frame_transmission_timestamp = first_frame_transmission_timestamp

    @property
    def original_transfer_timestamp(self) -> Timestamp:
        return self._original_transfer_timestamp

    @property
    def first_frame_transmission_timestamp(self) -> Timestamp:
        return self._first_frame_transmission_timestamp


@dataclasses.dataclass(frozen=True)
class _PendingFeedbackKey:
    compiled_identifier: int
    transfer_id_modulus: int


# noinspection PyAbstractClass
class CANOutputSession(CANSession, pyuavcan.transport.OutputSession):
    """
    This is actually an abstract class, but its concrete inheritors are hidden from the API.
    The implementation is chosen according to the type of the session requested: broadcast or unicast.
    """

    def __init__(
        self,
        transport: pyuavcan.transport.can.CANTransport,
        send_handler: SendHandler,
        specifier: pyuavcan.transport.OutputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        finalizer: SessionFinalizer,
    ):
        """Use the factory method."""
        self._transport = transport
        self._send_handler = send_handler
        self._specifier = specifier
        self._payload_metadata = payload_metadata

        self._feedback_handler: typing.Optional[typing.Callable[[pyuavcan.transport.Feedback], None]] = None
        self._pending_feedback: typing.Dict[_PendingFeedbackKey, Timestamp] = {}

        self._statistics = pyuavcan.transport.SessionStatistics()

        super().__init__(finalizer=finalizer)

    def _handle_loopback_frame(self, timestamp: Timestamp, frame: UAVCANFrame) -> None:
        """
        This is a part of the transport-internal API. It's a public method despite the name because Python's
        visibility handling capabilities are limited. I guess we could define a private abstract base to
        handle this but it feels like too much work. Why can't we have protected visibility in Python?
        """
        if frame.start_of_transfer:
            key = _PendingFeedbackKey(compiled_identifier=frame.identifier, transfer_id_modulus=frame.transfer_id)
            try:
                original_timestamp = self._pending_feedback.pop(key)
            except KeyError:
                pass  # Do not log this because packet capture mode generates a lot of unattended loopback frames.
            else:
                if self._feedback_handler is not None:
                    feedback = CANFeedback(original_timestamp, timestamp)
                    try:
                        self._feedback_handler(feedback)
                    except Exception as ex:  # pragma: no cover
                        _logger.exception(
                            "%s: Unhandled exception in the output session feedback handler %s: %s",
                            self,
                            self._feedback_handler,
                            ex,
                        )

    @property
    def specifier(self) -> pyuavcan.transport.OutputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def enable_feedback(self, handler: typing.Callable[[pyuavcan.transport.Feedback], None]) -> None:
        self._feedback_handler = handler

    def disable_feedback(self) -> None:
        self._feedback_handler = None
        self._pending_feedback.clear()

    def sample_statistics(self) -> pyuavcan.transport.SessionStatistics:
        return copy.copy(self._statistics)

    def close(self) -> None:  # pylint: disable=useless-super-delegation
        super().close()

    async def _do_send(self, can_id: CANID, transfer: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
        self._raise_if_closed()

        # Decompose the outgoing transfer into individual CAN frames.
        compiled_identifier = can_id.compile(transfer.fragmented_payload)
        tid_mod = transfer.transfer_id % TRANSFER_ID_MODULO  # https://github.com/UAVCAN/pyuavcan/issues/120
        frames = list(
            serialize_transfer(
                compiled_identifier=compiled_identifier,
                transfer_id=tid_mod,
                fragmented_payload=transfer.fragmented_payload,
                max_frame_payload_bytes=self._transport.protocol_parameters.mtu,
            )
        )

        # Ensure we're not trying to emit a multi-frame anonymous transfer - that's illegal.
        if can_id.source_node_id is None and len(frames) > 1:
            raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                f"Anonymous nodes cannot emit multi-frame transfers. CANID: {can_id}, transfer: {transfer}"
            )

        # If a loopback was requested, register it in the pending loopback registry.
        loopback_first_frame = self._feedback_handler is not None
        if loopback_first_frame:
            key = _PendingFeedbackKey(compiled_identifier=compiled_identifier, transfer_id_modulus=tid_mod)
            try:
                old = self._pending_feedback[key]
            except KeyError:
                pass
            else:
                self._statistics.errors += 1
                _logger.warning("%s: Overriding old feedback entry %s at key %s", self, old, key)
            self._pending_feedback[key] = transfer.timestamp

        # Emit the frames and update the statistical counters.
        try:
            transaction = SendTransaction(
                frames=frames, loopback_first=loopback_first_frame, monotonic_deadline=monotonic_deadline
            )
            if await self._send_handler(transaction):
                self._statistics.transfers += 1
                self._statistics.frames += len(frames)
                self._statistics.payload_bytes += sum(map(len, transfer.fragmented_payload))  # Session level
                return True
            self._statistics.drops += len(frames)
            return False
        except Exception:
            self._statistics.errors += 1
            raise


class BroadcastCANOutputSession(CANOutputSession):
    def __init__(
        self,
        specifier: pyuavcan.transport.OutputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        transport: pyuavcan.transport.can.CANTransport,
        send_handler: SendHandler,
        finalizer: SessionFinalizer,
    ):
        """Use the factory method."""
        assert specifier.remote_node_id is None, "Internal protocol violation: expected broadcast"
        if not isinstance(specifier.data_specifier, pyuavcan.transport.MessageDataSpecifier):
            raise pyuavcan.transport.UnsupportedSessionConfigurationError(
                f"This transport does not support broadcast outputs for {specifier.data_specifier}"
            )
        self._subject_id = specifier.data_specifier.subject_id

        super().__init__(
            transport=transport,
            send_handler=send_handler,
            specifier=specifier,
            payload_metadata=payload_metadata,
            finalizer=finalizer,
        )

    async def send(self, transfer: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
        can_id = MessageCANID(
            priority=transfer.priority,
            subject_id=self._subject_id,
            source_node_id=self._transport.local_node_id,  # May be anonymous
        )
        return await self._do_send(can_id, transfer, monotonic_deadline)


class UnicastCANOutputSession(CANOutputSession):
    def __init__(
        self,
        specifier: pyuavcan.transport.OutputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        transport: pyuavcan.transport.can.CANTransport,
        send_handler: SendHandler,
        finalizer: SessionFinalizer,
    ):
        """Use the factory method."""
        assert isinstance(specifier.remote_node_id, int), "Internal protocol violation: expected unicast"
        self._destination_node_id = int(specifier.remote_node_id)
        if not isinstance(specifier.data_specifier, pyuavcan.transport.ServiceDataSpecifier):
            raise pyuavcan.transport.UnsupportedSessionConfigurationError(
                f"This transport does not support unicast outputs for {specifier.data_specifier}"
            )
        if transport.local_node_id is None:
            raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                "Cannot emit service transfers because the local node is anonymous (does not have a node-ID)"
            )
        self._service_id = specifier.data_specifier.service_id
        self._request_not_response = (
            specifier.data_specifier.role == pyuavcan.transport.ServiceDataSpecifier.Role.REQUEST
        )

        super().__init__(
            transport=transport,
            send_handler=send_handler,
            specifier=specifier,
            payload_metadata=payload_metadata,
            finalizer=finalizer,
        )

    async def send(self, transfer: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
        source_node_id = self._transport.local_node_id
        assert source_node_id is not None, "Internal logic error"
        can_id = ServiceCANID(
            priority=transfer.priority,
            service_id=self._service_id,
            request_not_response=self._request_not_response,
            source_node_id=source_node_id,
            destination_node_id=self._destination_node_id,
        )
        return await self._do_send(can_id, transfer, monotonic_deadline)



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import sys
import time
from typing import Callable, Any, Optional, cast, Sequence
import ctypes
import socket
import logging
import threading
import dataclasses
import pyuavcan
from pyuavcan.transport import Timestamp


_logger = logging.getLogger(__name__)


class LinkLayerError(pyuavcan.transport.TransportError):
    pass


class LinkLayerCaptureError(LinkLayerError):
    pass


@dataclasses.dataclass(frozen=True)
class LinkLayerPacket:
    """
    OSI L2 packet representation.
    The addresses are represented here in the link-native byte order (big endian for Ethernet).
    """

    protocol: socket.AddressFamily
    """
    The protocol encapsulated inside this link-layer packet; e.g., IPv6.
    """

    source: memoryview
    destination: memoryview
    """
    Link-layer addresses, if applicable. Empty if not supported by the link layer.
    """

    payload: memoryview
    """
    The packet of the specified protocol.
    """

    def __repr__(self) -> str:
        """
        The repr displays only the first 100 bytes of the payload.
        If the payload is longer, its string representation is appended with an ellipsis.
        """
        limit = 100
        if len(self.payload) <= limit:
            pld = bytes(self.payload).hex()
        else:
            pld = bytes(self.payload[:limit]).hex() + "..."
        return pyuavcan.util.repr_attributes(
            self,
            protocol=str(self.protocol),
            source=bytes(self.source).hex(),
            destination=bytes(self.destination).hex(),
            payload=pld,
        )


@dataclasses.dataclass(frozen=True)
class LinkLayerCapture:
    timestamp: Timestamp
    packet: LinkLayerPacket
    device_name: str
    # Do we also need to report the link layer type here?


class LinkLayerSniffer:
    """
    This wrapper is intended to insulate the rest of the transport implementation from the specifics of the
    libpcap wrapper implementation (there are dozens of different wrappers out there).
    Observe that anything libpcap-related shall not be imported outside of these methods because we only require
    this dependency if protocol sniffing capability is needed.
    Regular use of the library should be possible without libpcap installed.

    Once a new instance is constructed, it is launched immediately.
    Execution is carried out in a background daemon thread pool.
    It is required to call :meth:`close` when done, which will hint the worker threads to terminate soon.

    If a new network device is added or re-initialized while the sniffer is running, it will not be recognized.
    Removal or a re-configuration of a device while the sniffer is running may cause it to fail,
    which will be logged from the worker threads.

    Should a worker thread encounter an error (e.g., if the device becomes unavailable), its capture context
    is closed automatically and then the thread is terminated.
    Such occurrences are logged at the CRITICAL severity level.

    - https://www.tcpdump.org/manpages/pcap.3pcap.html
    - https://github.com/karpierz/libpcap/blob/master/tests/capturetest.py
    """

    def __init__(self, filter_expression: str, callback: Callable[[LinkLayerCapture], None]) -> None:
        """
        :param filter_expression: The standard pcap filter expression;
            see https://www.tcpdump.org/manpages/pcap-filter.7.html.
            Use Wireshark for testing filter expressions.

        :param callback: This callback will be invoked once whenever a packet is captured with a single argument
            of type :class:`LinkLayerCapture`.
            Notice an important detail: the sniffer takes care of managing the link layer packets.
            The user does not need to care which type of data link layer encapsulation is used:
            it could be Ethernet, IEEE 802.15.4, or whatever.
            The application always gets a high-level view of the data with the link-layer specifics abstracted away.
            This function may be invoked directly from a worker thread, so be sure to apply synchronization.
        """
        self._filter_expr = str(filter_expression)
        self._callback = callback
        self._keep_going = True
        self._workers: list[threading.Thread] = []
        try:
            dev_names = _find_devices()
            _logger.debug("Capturable network devices: %s", dev_names)
            caps = _capture_all(dev_names, filter_expression)
        except PermissionError:
            if sys.platform.startswith("linux"):
                suggestion = f'Run this:\nsudo setcap cap_net_raw+eip "$(readlink -f {sys.executable})"'
            elif sys.platform.startswith("win"):
                suggestion = "Make sure you have Npcap installed and configured properly: https://nmap.org/npcap"
            else:
                suggestion = ""
            raise PermissionError(
                f"You need special privileges to perform low-level network packet capture (sniffing). {suggestion}"
            ) from None
        if not caps:
            raise LinkLayerCaptureError(
                f"There are no devices available for packet capture at the moment. Evaluated candidates: {dev_names}"
            )
        self._workers = [
            threading.Thread(target=self._thread_worker, name=f"pcap_{name}", args=(name, pd, decoder), daemon=True)
            for name, pd, decoder in caps
        ]
        for w in self._workers:
            w.start()
        assert len(self._workers) > 0

    @property
    def is_stable(self) -> bool:
        """
        True if all devices detected during the initial configuration are still being captured from.
        If at least one of them failed (e.g., due to a system reconfiguration), this value would be false.
        """
        assert len(self._workers) > 0
        return all(x.is_alive() for x in self._workers)

    def close(self) -> None:
        """
        After closing the callback reference is immediately destroyed to prevent the receiver from being kept alive
        by the not-yet-terminated worker threads and to prevent residual packets from generating spurious events.
        """
        self._keep_going = False
        self._callback = lambda *_: None
        # This is not a great solution, honestly. Consider improving it later.
        # Currently we just unbind the callback from the user-supplied destination and mark that the threads should
        # terminate. The sniffer is then left in a locked-in state, where it may keep performing some no-longer-useful
        # activities in the background, but they remain invisible to the outside world. Eventually, the instance will
        # be disposed after the last worker is terminated, but we should make it more deterministic.

    def _thread_worker(self, name: str, pd: object, decoder: PacketDecoder) -> None:
        import libpcap as pcap  # type: ignore

        assert isinstance(pd, ctypes.POINTER(pcap.pcap_t))
        try:
            _logger.debug("%r: Worker thread for %r is started: %s", self, name, threading.current_thread())

            # noinspection PyTypeChecker
            @pcap.pcap_handler  # type: ignore
            def proxy(_: object, header: ctypes.Structure, packet: Any) -> None:
                # Parse the header, extract the timestamp and the packet length.
                header = header.contents
                ts_ns = (header.ts.tv_sec * 1_000_000 + header.ts.tv_usec) * 1000
                ts = Timestamp(system_ns=ts_ns, monotonic_ns=time.monotonic_ns())
                length, real_length = header.caplen, header.len
                _logger.debug("%r: CAPTURED PACKET ts=%s dev=%r len=%d bytes", self, ts, name, length)
                if real_length != length:
                    # In theory, this should never occur because we use a huge capture buffer.
                    # On Windows, however, when using Npcap v0.96, the captured length is (always?) reported to be
                    # 32 bytes shorter than the real length, despite the fact that the packet is not truncated.
                    _logger.debug(
                        "%r: Length mismatch in a packet captured from %r: real %r bytes, captured %r bytes",
                        self,
                        name,
                        real_length,
                        length,
                    )
                # Create a copy of the payload. This is required per the libpcap API contract -- it says that the
                # memory is invalidated upon return from the callback.
                packet = memoryview(ctypes.cast(packet, ctypes.POINTER(ctypes.c_ubyte * length))[0]).tobytes()
                llp = decoder(memoryview(packet))
                if llp is None:
                    if _logger.isEnabledFor(logging.INFO):
                        _logger.info(
                            "%r: Link-layer packet of %d bytes captured from %r at %s could not be parsed. "
                            "The header is: %s",
                            self,
                            len(packet),
                            name,
                            ts,
                            packet[:32].hex(),
                        )
                else:
                    self._callback(LinkLayerCapture(timestamp=ts, packet=llp, device_name=name))

            packets_per_batch = 100
            while self._keep_going:
                err = pcap.dispatch(pd, packets_per_batch, proxy, ctypes.POINTER(ctypes.c_ubyte)())
                if err < 0:  # Negative values represent errors, otherwise it's the number of packets processed.
                    if self._keep_going:
                        _logger.critical(
                            "%r: Worker thread for %r has failed with error %s; %s",
                            self,
                            name,
                            err,
                            pcap.geterr(pd).decode(),
                        )
                    else:
                        _logger.debug(
                            "%r: Error %r in worker thread for %r ignored because it is commanded to stop",
                            self,
                            err,
                            name,
                        )
                    break
        except Exception as ex:
            _logger.exception("%r: Unhandled exception in worker thread for %r; stopping: %r", self, name, ex)
        finally:
            # BEWARE: pcap_close() is not idempotent! Second close causes a heap corruption. *sigh*
            pcap.close(pd)
        _logger.debug("%r: Worker thread for %r is being terminated", self, name)

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self,
            filter_expression=repr(self._filter_expr),
            num_devices=len(self._workers),
            num_devices_active=len(list(x.is_alive() for x in self._workers)),
        )


PacketEncoder = Callable[["LinkLayerPacket"], Optional[memoryview]]
PacketDecoder = Callable[[memoryview], Optional["LinkLayerPacket"]]


def _get_codecs() -> dict[int, tuple[PacketEncoder, PacketDecoder]]:
    """
    A factory of paired encode/decode functions that are used for building and parsing link-layer packets.
    The pairs are organized into a dict where the key is the data link type code from libpcap;
    see https://www.tcpdump.org/linktypes.html.
    The dict is ordered such that the recommended data link types come first.
    This is useful when setting up packet capture if the adapter supports multiple link layer formats.

    The encoder returns None if the encapsulated protocol is not supported by the selected link layer.
    The decoder returns None if the packet is not valid or the encapsulated protocol is not supported.
    """
    import libpcap as pcap
    from socket import AddressFamily

    def get_ethernet() -> tuple[PacketEncoder, PacketDecoder]:
        # https://en.wikipedia.org/wiki/EtherType
        af_to_ethertype = {
            AddressFamily.AF_INET: 0x0800,
            AddressFamily.AF_INET6: 0x86DD,
        }
        ethertype_to_af = {v: k for k, v in af_to_ethertype.items()}

        def enc(p: LinkLayerPacket) -> Optional[memoryview]:
            try:
                return memoryview(
                    b"".join(
                        (
                            bytes(p.source).rjust(6, b"\x00")[:6],
                            bytes(p.destination).rjust(6, b"\x00")[:6],
                            af_to_ethertype[p.protocol].to_bytes(2, "big"),
                            p.payload,
                        )
                    )
                )
            except LookupError:
                return None

        def dec(p: memoryview) -> Optional[LinkLayerPacket]:
            if len(p) < 14:
                return None
            src = p[0:6]
            dst = p[6:12]
            ethertype = int.from_bytes(p[12:14], "big")
            try:
                protocol = ethertype_to_af[ethertype]
            except LookupError:
                return None
            return LinkLayerPacket(protocol=protocol, source=src, destination=dst, payload=p[14:])

        return enc, dec

    def get_loopback(byte_order: str) -> tuple[PacketEncoder, PacketDecoder]:
        # DLT_NULL is used by the Windows loopback interface. Info: https://wiki.wireshark.org/NullLoopback
        # The source and destination addresses are not representable in this data link layer.
        def enc(p: LinkLayerPacket) -> Optional[memoryview]:
            return memoryview(b"".join((p.protocol.to_bytes(4, byte_order), p.payload)))  # type: ignore

        def dec(p: memoryview) -> Optional[LinkLayerPacket]:
            if len(p) < 4:
                return None
            try:
                protocol = AddressFamily(int.from_bytes(p[0:4], byte_order))  # type: ignore
            except ValueError:
                return None
            empty = memoryview(b"")
            return LinkLayerPacket(protocol=protocol, source=empty, destination=empty, payload=p[4:])

        return enc, dec

    # The output is ORDERED, best option first.
    return {
        pcap.DLT_EN10MB: get_ethernet(),
        pcap.DLT_LOOP: get_loopback("big"),
        pcap.DLT_NULL: get_loopback(sys.byteorder),
    }


def _find_devices() -> list[str]:
    """
    Returns a list of local network devices that can be captured from.
    Raises a PermissionError if the user is suspected to lack the privileges necessary for capture.

    We used to filter the devices by address family, but it turned out to be a dysfunctional solution because
    a device does not necessarily have to have an address in a particular family to be able to capture packets
    of that kind. For instance, on Windows, a virtual network adapter may have no addresses while still being
    able to capture packets.
    """
    import libpcap as pcap

    err_buf = ctypes.create_string_buffer(pcap.PCAP_ERRBUF_SIZE)
    devices = ctypes.POINTER(pcap.pcap_if_t)()
    if pcap.findalldevs(ctypes.byref(devices), err_buf) != 0:
        raise LinkLayerError(f"Could not list network devices: {err_buf.value.decode()}")
    if not devices:
        # This may seem odd, but libpcap returns an empty list if the user is not allowed to perform capture.
        # This is documented in the API docs as follows:
        #   Note that there may be network devices that cannot be opened by the process calling pcap_findalldevs(),
        #   because, for example, that process does not have sufficient privileges to open them for capturing;
        #   if so, those devices will not appear on the list.
        raise PermissionError("No capturable devices have been found. Do you have the required privileges?")
    dev_names: list[str] = []
    d = cast(ctypes.Structure, devices)
    while d:
        d = d.contents
        name = d.name.decode()
        if name != "any":
            dev_names.append(name)
        else:
            _logger.debug("Synthetic device %r does not support promiscuous mode, skipping", name)
        d = d.next
    pcap.freealldevs(devices)
    return dev_names


def _capture_all(device_names: list[str], filter_expression: str) -> list[tuple[str, object, PacketDecoder]]:
    """
    Begin capture on all devices in promiscuous mode.
    We can't use "any" because libpcap does not support promiscuous mode with it, as stated in the docs and here:
    https://github.com/the-tcpdump-group/libpcap/blob/bcca74d2713dc9c0a27992102c469f77bdd8dd1f/pcap-linux.c#L2522.
    It shouldn't be a problem because we have our filter expression that is expected to be highly efficient.
    Devices whose ifaces are down or that are not usable for other valid reasons will be silently filtered out here.
    """
    import libpcap as pcap

    codecs = _get_codecs()
    caps: list[tuple[str, object, PacketDecoder]] = []
    try:
        for name in device_names:
            pd = _capture_single_device(name, filter_expression, list(codecs.keys()))
            if pd is None:
                _logger.info("Could not set up capture on %r", name)
                continue
            data_link_type = pcap.datalink(pd)
            try:
                _, dec = codecs[data_link_type]
            except LookupError:
                # This is where we filter out devices that certainly have no relevance, like CAN adapters.
                pcap.close(pd)
                _logger.info(
                    "Device %r will not be used for packet capture because its data link layer type=%r "
                    "is not supported by this library. Either the device is irrelevant, "
                    "or the library needs to be extended to support this link layer protocol.",
                    name,
                    data_link_type,
                )
            else:
                caps.append((name, pd, dec))
    except Exception:
        for _, c, _ in caps:
            pcap.close(c)
        raise
    _logger.info(
        "Capture sessions with filter %r have been set up on: %s", filter_expression, list(n for n, _, _ in caps)
    )
    return caps


def _capture_single_device(device: str, filter_expression: str, data_link_hints: Sequence[int]) -> Optional[object]:
    """
    Returns None if the interface managed by this device is not up or if it cannot be captured from for other reasons.
    On GNU/Linux, some virtual devices (like netfilter devices) can only be accessed by a superuser.

    The function will configure libpcap to use the first supported data link type from the list.
    If none of the specified data link types are supported, a log message is emitted but no error is raised.
    The available link types are listed in https://www.tcpdump.org/linktypes.html.
    """
    import libpcap as pcap

    def status_to_str(error_code: int) -> str:
        """
        Some libpcap-compatible libraries (e.g., WinPCap) do not have this function, so we have to define a fallback.
        """
        try:
            return str(pcap.statustostr(error_code).decode())
        except AttributeError:  # pragma: no cover
            return f"[error {error_code}]"

    # This is helpful: https://github.com/karpierz/libpcap/blob/master/tests/capturetest.py
    err_buf = ctypes.create_string_buffer(pcap.PCAP_ERRBUF_SIZE)
    pd = pcap.create(device.encode(), err_buf)
    if pd is None:
        raise LinkLayerCaptureError(f"Could not instantiate pcap_t for {device!r}: {err_buf.value.decode()}")
    try:
        # Non-fatal errors are intentionally logged at a low severity level to not disturb the user unnecessarily.
        err = pcap.set_snaplen(pd, _SNAPSHOT_LENGTH)
        if err != 0:
            _logger.info("Could not set snapshot length for %r: %r", device, status_to_str(err))

        err = pcap.set_timeout(pd, int(_BUFFER_TIMEOUT * 1e3))
        if err != 0:
            _logger.info("Could not set timeout for %r: %r", device, status_to_str(err))

        err = pcap.set_promisc(pd, 1)
        if err != 0:
            _logger.info("Could not enable promiscuous mode for %r: %r", device, status_to_str(err))

        err = pcap.activate(pd)
        if err in (pcap.PCAP_ERROR_PERM_DENIED, pcap.PCAP_ERROR_PROMISC_PERM_DENIED):
            raise PermissionError(f"Capture is not permitted on {device!r}: {status_to_str(err)}")
        if err == pcap.PCAP_ERROR_IFACE_NOT_UP:
            _logger.debug("Device %r is not capturable because the iface is not up. %s", device, status_to_str(err))
            pcap.close(pd)
            return None
        if err < 0:
            _logger.info(
                "Could not activate capture on %r: %s; %s", device, status_to_str(err), pcap.geterr(pd).decode()
            )
            pcap.close(pd)
            return None
        if err > 0:
            _logger.info(
                "Capture on %r started successfully, but libpcap reported a warning: %s", device, status_to_str(err)
            )

        # https://www.tcpdump.org/manpages/pcap_set_datalink.3pcap.html
        for dlt in data_link_hints:
            err = pcap.set_datalink(pd, dlt)
            if err == 0:
                _logger.debug("Device %r is configured to use the data link type %r", device, dlt)
                break
        else:
            _logger.debug(
                "Device %r supports none of the following data link types: %r. Last error was: %s",
                device,
                list(data_link_hints),
                pcap.geterr(pd).decode(),
            )
            return None

        # https://www.tcpdump.org/manpages/pcap_compile.3pcap.html
        code = pcap.bpf_program()  # This memory needs to be freed when closed. Fix it later.
        err = pcap.compile(pd, ctypes.byref(code), filter_expression.encode(), 1, pcap.PCAP_NETMASK_UNKNOWN)
        if err != 0:
            raise LinkLayerCaptureError(
                f"Could not compile filter expression {filter_expression!r}: {status_to_str(err)}; "
                f"{pcap.geterr(pd).decode()}"
            )
        err = pcap.setfilter(pd, ctypes.byref(code))
        if err != 0:
            raise LinkLayerCaptureError(f"Could not install filter: {status_to_str(err)}; {pcap.geterr(pd).decode()}")
    except Exception:
        pcap.close(pd)
        raise
    return cast(object, pd)


_SNAPSHOT_LENGTH = 65535
"""
The doc says: "A snapshot length of 65535 should be sufficient, on most if not all networks,
to capture all the data available from the packet."
"""

_BUFFER_TIMEOUT = 0.005
"""
See "packet buffer timeout" in https://www.tcpdump.org/manpages/pcap.3pcap.html.
This value should be sensible for any kind of real-time monitoring application.
"""


def _apply_windows_workarounds() -> None:  # pragma: no cover
    import os
    import pathlib
    import importlib.util

    # This is a Windows Server-specific workaround for this libpcap issue: https://github.com/karpierz/libpcap/issues/7
    # tl;dr: It works on desktop Windows 8/10, but Windows Server 2019 is unable to find "wpcap.dll" unless the
    # DLL search path is specified manually via PATH. The workaround is valid per libpcap==1.10.0b15.
    # Later versions of libpcap may not require it, so please consider removing it in the future.
    spec = importlib.util.find_spec("libpcap")
    if spec and spec.origin:
        is_64_bit = sys.maxsize.bit_length() > 32
        libpcap_dir = pathlib.Path(spec.origin).parent
        dll_path = libpcap_dir / "_platform" / "_windows" / ("x64" if is_64_bit else "x86") / "wpcap"
        os.environ["PATH"] += os.pathsep + str(dll_path)


if sys.platform.startswith("win"):  # pragma: no cover
    _apply_windows_workarounds()



# Copyright (C) 2021  UAVCAN Consortium  <uavcan.org>
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from typing import Tuple, Optional, Callable, Dict, Iterator, Union
import logging
from . import Entry, BackendError, Backend, Value


__all__ = ["DynamicBackend"]


class DynamicBackend(Backend):
    """
    Register backend where register access is delegated to external getters and setters.
    It does not store values internally.
    Exceptions raised by getters/setters are wrapped into :class:`BackendError`.

    Create new registers and change value of existing ones using :meth:`__setitem__`.

    >>> from pyuavcan.application.register import Bit
    >>> b = DynamicBackend()
    >>> b.persistent
    False
    >>> b.get("foo") is None
    True
    >>> b.index(0) is None
    True
    >>> foo = Value(bit=Bit([True, False, True]))
    >>> def set_foo(v: Value):
    ...     global foo
    ...     foo = v
    >>> b["foo"] = (lambda: foo), set_foo   # Create new mutable register.
    >>> b["foo"].mutable
    True
    >>> list(b["foo"].value.bit.value)
    [True, False, True]
    >>> b["foo"] = Value(bit=Bit([False, True, True]))  # Set new value.
    >>> list(b["foo"].value.bit.value)
    [False, True, True]
    >>> b["foo"] = lambda: foo  # Replace register with a new one that is now immutable.
    >>> b["foo"] = Value(bit=Bit([False, False, False]))    # Value cannot be changed.
    >>> list(b["foo"].value.bit.value)
    [False, True, True]
    >>> list(b)
    ['foo']
    >>> del b["foo"]
    >>> list(b)
    []
    """

    def __init__(self) -> None:
        self._reg: Dict[str, GetSetPair] = {}  # This dict is always sorted lexicographically by key!
        super().__init__()

    @property
    def location(self) -> str:
        """This is a stub."""
        return ""

    @property
    def persistent(self) -> bool:
        """Always false."""
        return False

    def close(self) -> None:
        """Clears all registered registers."""
        self._reg.clear()

    def index(self, index: int) -> Optional[str]:
        try:
            return list(self)[index]
        except LookupError:
            return None

    def __getitem__(self, key: str) -> Entry:
        getter, setter = self._reg[key]
        try:
            value = getter()
        except Exception as ex:
            raise BackendError(f"Unhandled exception in getter for {key!r}: {ex}") from ex
        e = Entry(value, mutable=setter is not None)
        _logger.debug("%r: Get %r -> %r", self, key, e)
        return e

    def __setitem__(
        self,
        key: str,
        value: Union[
            Entry,
            Value,
            Callable[[], Value],
            Tuple[Callable[[], Value], Callable[[Value], None]],
        ],
    ) -> None:
        """
        :param key: The register name.

        :param value:
            - If this is an instance of :class:`Entry` or :class:`Value`, and the referenced register is mutable,
              its setter is invoked with the supplied instance of :class:`Value`
              (if :class:`Entry` is given, the value is extracted from there and the mutability flag is ignored).
              If the register is immutable, nothing is done.
              The caller is required to ensure that the type is acceptable.

            - If this is a single callable, a new immutable register is defined (existing registers overwritten).

            - If this is a tuple of two callables, a new mutable register is defined (existing registers overwritten).
        """
        if isinstance(value, Entry):
            value = value.value

        if isinstance(value, Value):
            try:
                _, setter = self._reg[key]
            except LookupError:
                setter = None
            if setter is not None:
                _logger.debug("%r: Set %r <- %r", self, key, value)
                try:
                    setter(value)
                except Exception as ex:
                    raise BackendError(f"Unhandled exception in setter for {key!r}: {ex}") from ex
            else:
                _logger.debug("%r: Set %r not supported", self, key)
        else:
            if callable(value):
                getter, setter = value, None
            elif isinstance(value, tuple) and len(value) == 2 and all(map(callable, value)):
                getter, setter = value
            else:  # pragma: no cover
                raise TypeError(f"Invalid argument: {value!r}")
            items = list(self._reg.items())
            items.append((key, (getter, setter)))
            self._reg = dict(sorted(items, key=lambda x: x[0]))

    def __delitem__(self, key: str) -> None:
        _logger.debug("%r: Delete %r", self, key)
        del self._reg[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._reg)

    def __len__(self) -> int:
        return len(self._reg)


GetSetPair = Tuple[
    Callable[[], Value],
    Optional[Callable[[Value], None]],
]

_logger = logging.getLogger(__name__)


def _unittest_dyn() -> None:
    from uavcan.primitive import String_1 as String

    b = DynamicBackend()
    assert not b.persistent
    assert len(b) == 0
    assert list(b.keys()) == []
    assert b.get("foo") is None
    assert b.index(0) is None

    bar = Value(string=String())

    def set_bar(v: Value) -> None:
        nonlocal bar
        bar = v

    b["foo"] = lambda: Value(string=String("Hello"))
    b["bar"] = lambda: bar, set_bar
    assert len(b) == 2
    assert list(b.keys()) == ["bar", "foo"]
    assert b.index(0) == "bar"
    assert b.index(1) == "foo"
    assert b.index(2) is None

    e = b.get("foo")
    assert e
    assert not e.mutable
    assert e.value.string
    assert e.value.string.value.tobytes().decode() == "Hello"

    e = b.get("bar")
    assert e
    assert e.mutable
    assert e.value.string
    assert e.value.string.value.tobytes().decode() == ""

    b["foo"] = Value(string=String("world"))
    b["bar"] = Entry(Value(string=String("world")), mutable=False)  # Flag ignored

    e = b.get("foo")
    assert e
    assert not e.mutable
    assert e.value.string
    assert e.value.string.value.tobytes().decode() == "Hello"

    e = b.get("bar")
    assert e
    assert e.mutable
    assert e.value.string
    assert e.value.string.value.tobytes().decode() == "world"

    del b["foo"]
    assert len(b) == 1
    assert list(b.keys()) == ["bar"]

    b.close()
    assert len(b) == 0



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import enum
import typing
import logging
import pyuavcan
from pyuavcan.transport import Timestamp, Priority, TransferFrom
from ._frame import Frame
from ._common import TransferCRC


_logger = logging.getLogger(__name__)


_CRC_SIZE_BYTES = len(TransferCRC().value_as_bytes)


class TransferReassembler:
    """
    Multi-frame transfer reassembly logic is arguably the most complex part of any UAVCAN transport implementation.
    This class implements a highly transport-agnostic transfer reassembly state machine designed for use
    with high-overhead transports, such as UDP, Serial, IEEE 802.15.4, etc.
    Any transport whose frame dataclass implementation derives from :class:`Frame` can use this class.

    Out-of-order frame reception is supported, and therefore the reassembler can be used with
    redundant interfaces directly, without preliminary frame deduplication procedures or explicit
    interface index assignment, provided that all involved redundant interfaces share the same MTU setting.
    OOO support includes edge cases where the first frame of a transfer is not received first and/or the last
    frame is not received last.

    OOO is required for frame-level modular transport redundancy (more than one transport operating concurrently)
    and temporal transfer redundancy (every transfer repeated several times to mitigate frame loss).
    The necessity of OOO is due to the fact that frames sourced concurrently from multiple transport interfaces
    and/or frames of a temporally redundant transfer where some of the frames are lost
    result in an out-of-order arrival of the frames.
    Additionally, various non-vehicular and/or non-mission-critical networks
    (such as conventional IP networks) may deliver frames out-of-order even without redundancy.

    Distantly relevant discussion: https://github.com/UAVCAN/specification/issues/8.

    A multi-frame transfer shall not contain frames with empty payload.
    """

    class Error(enum.Enum):
        """
        Error states that the transfer reassembly state machine may encounter.
        Whenever an error is encountered, the corresponding error counter is incremented by one,
        and a verbose report is dumped into the log at the DEBUG level.
        """

        MULTIFRAME_MISSING_FRAMES = enum.auto()
        """
        New transfer started before the old one could be completed. Old transfer discarded.
        """

        MULTIFRAME_INTEGRITY_ERROR = enum.auto()
        """
        A reassembled multi-frame transfer payload did not pass integrity checks. Transfer discarded.
        """

        MULTIFRAME_EMPTY_FRAME = enum.auto()
        """
        A frame without payload received as part of a multiframe transfer (not permitted by Specification).
        Only single-frame transfers can have empty payload.
        """

        MULTIFRAME_EOT_MISPLACED = enum.auto()
        """
        The end-of-transfer flag is set in a frame with index N,
        but the transfer contains at least one frame with index > N. Transfer discarded.
        """

        MULTIFRAME_EOT_INCONSISTENT = enum.auto()
        """
        The end-of-transfer flag is set in frames with indexes N and M, where N != M. Transfer discarded.
        """

    def __init__(
        self,
        source_node_id: int,
        extent_bytes: int,
        on_error_callback: typing.Callable[[TransferReassembler.Error], None],
    ):
        """
        :param source_node_id: The remote node-ID whose transfers this instance will be listening for.
            Anonymous transfers cannot be multi-frame transfers, so they are to be accepted as-is without any
            reassembly activities.

        :param extent_bytes: The maximum number of payload bytes per transfer.
            Payload that exceeds this size limit may be implicitly truncated (in the Specification this behavior
            is described as "implicit truncation rule").
            This value can be derived from the corresponding DSDL definition.

        :param on_error_callback: The callback is invoked whenever an error is detected.
            This is intended for diagnostic purposes only; the error information is not actionable.
            The error is logged by the caller at the DEBUG verbosity level together with reassembly context info.
        """
        # Constant configuration.
        self._source_node_id = int(source_node_id)
        self._extent_bytes = int(extent_bytes)
        self._on_error_callback = on_error_callback
        if self._source_node_id < 0 or self._extent_bytes < 0 or not callable(self._on_error_callback):
            raise ValueError("Invalid parameters")

        # Internal state.
        self._payloads: typing.List[memoryview] = []  # Payload fragments from the received frames.
        self._max_index: typing.Optional[int] = None  # Max frame index in transfer, None if unknown.
        self._timestamp = Timestamp(0, 0)  # First frame timestamp.
        self._transfer_id = 0  # Transfer-ID of the current transfer.

    def process_frame(
        self, timestamp: Timestamp, frame: Frame, transfer_id_timeout: float
    ) -> typing.Optional[TransferFrom]:
        """
        Updates the transfer reassembly state machine with the new frame.

        :param timestamp: The reception timestamp from the transport layer.
        :param frame: The new frame.
        :param transfer_id_timeout: The current value of the transfer-ID timeout.
        :return: A new transfer if the new frame completed one. None if the new frame did not complete a transfer.
        :raises: Nothing.
        """
        # DROP MALFORMED FRAMES. A multi-frame transfer cannot contain frames with no payload.
        if not (frame.index == 0 and frame.end_of_transfer) and not frame.payload:
            self._on_error_callback(self.Error.MULTIFRAME_EMPTY_FRAME)
            return None

        # DETECT NEW TRANSFERS. Either a newer TID or TID-timeout is reached.
        if (
            frame.transfer_id > self._transfer_id
            or timestamp.monotonic - self._timestamp.monotonic > transfer_id_timeout
        ):
            self._restart(
                timestamp, frame.transfer_id, self.Error.MULTIFRAME_MISSING_FRAMES if self._payloads else None
            )

        # DROP FRAMES FROM NON-MATCHING TRANSFERS. E.g., duplicates. This is not an error.
        if frame.transfer_id < self._transfer_id:
            return None
        assert frame.transfer_id == self._transfer_id

        # DETERMINE MAX FRAME INDEX FOR THIS TRANSFER. Frame N with EOT, then frame M with EOT, where N != M.
        if frame.end_of_transfer:
            if self._max_index is not None and self._max_index != frame.index:
                self._restart(timestamp, frame.transfer_id + 1, self.Error.MULTIFRAME_EOT_INCONSISTENT)
                return None
            assert self._max_index is None or self._max_index == frame.index
            self._max_index = frame.index

        # DETECT UNEXPECTED FRAMES PAST THE END OF TRANSFER. If EOT is set on index N, then indexes > N are invalid.
        if self._max_index is not None and max(frame.index, len(self._payloads) - 1) > self._max_index:
            self._restart(timestamp, frame.transfer_id + 1, self.Error.MULTIFRAME_EOT_MISPLACED)
            return None

        # ACCEPT THE PAYLOAD. Duplicates are accepted too, assuming they carry the same payload.
        # Implicit truncation is implemented by not limiting the maximum payload size.
        # Real truncation is hard to implement if frames are delivered out-of-order, although it's not impossible:
        # instead of storing actual payload fragments above the limit, we can store their CRCs.
        # When the last fragment is received, CRC of all fragments are then combined to validate the final transfer-CRC.
        # This method, however, requires knowledge of the MTU to determine which fragments will be above the limit.
        while len(self._payloads) <= frame.index:
            self._payloads.append(memoryview(b""))
        self._payloads[frame.index] = frame.payload

        # CHECK IF ALL FRAMES ARE RECEIVED. If not, simply wait for next frame.
        # Single-frame transfers with empty payload are legal.
        if self._max_index is None or (self._max_index > 0 and not all(self._payloads)):
            return None
        assert self._max_index is not None
        assert self._max_index == len(self._payloads) - 1
        assert all(self._payloads) if self._max_index > 0 else True

        # FINALIZE THE TRANSFER. All frames are received here.
        result = _validate_and_finalize_transfer(
            timestamp=self._timestamp,
            priority=frame.priority,
            transfer_id=frame.transfer_id,
            frame_payloads=self._payloads,
            source_node_id=self._source_node_id,
        )
        self._restart(
            timestamp, frame.transfer_id + 1, self.Error.MULTIFRAME_INTEGRITY_ERROR if result is None else None
        )
        if result is not None:
            # Late implicit truncation. Normally, it should be done on-the-fly, by not storing payload fragments
            # above the maximum expected size, but it is hard to combine with out-of-order frame acceptance.
            while result.fragmented_payload and sum(map(len, result.fragmented_payload[:-1])) > self._extent_bytes:
                # TODO: a minor refactoring is needed to avoid re-creating the transfer instance here.
                result = TransferFrom(
                    timestamp=result.timestamp,
                    priority=result.priority,
                    transfer_id=result.transfer_id,
                    fragmented_payload=result.fragmented_payload[:-1],
                    source_node_id=result.source_node_id,
                )
        return result

    @property
    def source_node_id(self) -> int:
        return self._source_node_id

    def _restart(
        self, timestamp: Timestamp, transfer_id: int, error: typing.Optional[TransferReassembler.Error] = None
    ) -> None:
        if error is not None:
            self._on_error_callback(error)
            if _logger.isEnabledFor(logging.DEBUG):  # pragma: no branch
                context = {
                    "ts": self._timestamp,
                    "tid": self._transfer_id,
                    "max_idx": self._max_index,
                    "payload": f"{len(list(x for x in self._payloads if x))}/{len(self._payloads)}",
                }
                _logger.debug(  # pylint: disable=logging-not-lazy
                    f"{self}: {error.name}: " + " ".join(f"{k}={v}" for k, v in context.items())
                )
        # The error must be processed before the state is reset because when the state is destroyed
        # the useful diagnostic information becomes unavailable.
        self._timestamp = timestamp
        self._transfer_id = transfer_id
        self._max_index = None
        self._payloads = []

    @property
    def _pure_payload_size_bytes(self) -> int:
        """May return a negative if the transfer is malformed."""
        size = sum(map(len, self._payloads))
        if len(self._payloads) > 1:
            size -= _CRC_SIZE_BYTES
        return size

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes_noexcept(
            self, source_node_id=self._source_node_id, extent_bytes=self._extent_bytes
        )

    @staticmethod
    def construct_anonymous_transfer(timestamp: Timestamp, frame: Frame) -> typing.Optional[TransferFrom]:
        """
        A minor helper that validates whether the frame is a valid anonymous transfer (it is if the index
        is zero and the end-of-transfer flag is set) and constructs a transfer instance if it is.
        Otherwise, returns None.
        Observe that this is a static method because anonymous transfers are fundamentally stateless.
        """
        if frame.single_frame_transfer:
            return TransferFrom(
                timestamp=timestamp,
                priority=frame.priority,
                transfer_id=frame.transfer_id,
                fragmented_payload=[frame.payload],
                source_node_id=None,
            )
        return None


def _validate_and_finalize_transfer(
    timestamp: Timestamp,
    priority: Priority,
    transfer_id: int,
    frame_payloads: typing.List[memoryview],
    source_node_id: int,
) -> typing.Optional[TransferFrom]:
    assert all(isinstance(x, memoryview) for x in frame_payloads)
    assert frame_payloads

    def package(fragmented_payload: typing.Sequence[memoryview]) -> TransferFrom:
        return TransferFrom(
            timestamp=timestamp,
            priority=priority,
            transfer_id=transfer_id,
            fragmented_payload=fragmented_payload,
            source_node_id=source_node_id,
        )

    if len(frame_payloads) > 1:
        size_ok = sum(map(len, frame_payloads)) > _CRC_SIZE_BYTES
        crc_ok = TransferCRC.new(*frame_payloads).check_residue()
        return package(_drop_crc(frame_payloads)) if size_ok and crc_ok else None
    return package(frame_payloads)


def _drop_crc(fragments: typing.List[memoryview]) -> typing.Sequence[memoryview]:
    remaining = _CRC_SIZE_BYTES
    while fragments and remaining > 0:
        if len(fragments[-1]) <= remaining:
            remaining -= len(fragments[-1])
            fragments.pop()
        else:
            fragments[-1] = fragments[-1][:-remaining]
            remaining = 0
    return fragments


# ----------------------------------------  TESTS BELOW THIS LINE  ----------------------------------------


def _unittest_transfer_reassembler() -> None:
    from pytest import raises

    src_nid = 1234
    prio = Priority.SLOW
    transfer_id_timeout = 1.0

    error_counters = {e: 0 for e in TransferReassembler.Error}

    def on_error_callback(error: TransferReassembler.Error) -> None:
        error_counters[error] += 1

    def mk_frame(
        transfer_id: int, index: int, end_of_transfer: bool, payload: typing.Union[bytes, memoryview]
    ) -> Frame:
        return Frame(
            priority=prio,
            transfer_id=transfer_id,
            index=index,
            end_of_transfer=end_of_transfer,
            payload=memoryview(payload),
        )

    def mk_transfer(
        timestamp: Timestamp, transfer_id: int, fragmented_payload: typing.Sequence[typing.Union[bytes, memoryview]]
    ) -> TransferFrom:
        return TransferFrom(
            timestamp=timestamp,
            priority=prio,
            transfer_id=transfer_id,
            fragmented_payload=list(map(memoryview, fragmented_payload)),
            source_node_id=src_nid,
        )

    def mk_ts(monotonic: float) -> Timestamp:
        monotonic_ns = round(monotonic * 1e9)
        return Timestamp(system_ns=monotonic_ns + 10 ** 12, monotonic_ns=monotonic_ns)

    with raises(ValueError):
        _ = TransferReassembler(source_node_id=-1, extent_bytes=100, on_error_callback=on_error_callback)

    with raises(ValueError):
        _ = TransferReassembler(source_node_id=0, extent_bytes=-1, on_error_callback=on_error_callback)

    ta = TransferReassembler(source_node_id=src_nid, extent_bytes=100, on_error_callback=on_error_callback)
    assert ta.source_node_id == src_nid

    def push(timestamp: Timestamp, frame: Frame) -> typing.Optional[TransferFrom]:
        return ta.process_frame(timestamp, frame, transfer_id_timeout=transfer_id_timeout)

    hedgehog = b"In the evenings, the little Hedgehog went to the Bear Cub to count stars."
    horse = b"He thought about the Horse: how was she doing there, in the fog?"

    # Valid single-frame transfer.
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=0, index=0, end_of_transfer=True, payload=hedgehog),
        )
        == mk_transfer(timestamp=mk_ts(1000.0), transfer_id=0, fragmented_payload=[hedgehog])
    )

    # Same transfer-ID; transfer ignored, no error registered.
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=0, index=0, end_of_transfer=True, payload=hedgehog),
        )
        is None
    )

    # Same transfer-ID, different EOT; transfer ignored, no error registered.
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=0, index=0, end_of_transfer=False, payload=hedgehog),
        )
        is None
    )

    # Valid multi-frame transfer.
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=2, index=0, end_of_transfer=False, payload=hedgehog[:50]),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(
                transfer_id=2,
                index=1,
                end_of_transfer=True,
                payload=hedgehog[50:] + TransferCRC.new(hedgehog).value_as_bytes,
            ),
        )
        == mk_transfer(timestamp=mk_ts(1000.0), transfer_id=2, fragmented_payload=[hedgehog[:50], hedgehog[50:]])
    )

    # Same as above, but the frame ordering is reversed.
    assert (
        push(
            mk_ts(1000.0),  # LAST FRAME
            mk_frame(transfer_id=10, index=2, end_of_transfer=True, payload=TransferCRC.new(hedgehog).value_as_bytes),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=10, index=1, end_of_transfer=False, payload=hedgehog[50:]),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),  # FIRST FRAME
            mk_frame(transfer_id=10, index=0, end_of_transfer=False, payload=hedgehog[:50]),
        )
        == mk_transfer(timestamp=mk_ts(1000.0), transfer_id=10, fragmented_payload=[hedgehog[:50], hedgehog[50:]])
    )

    # Same as above, but one frame is duplicated and one is ignored with old TID, plus an empty frame in the middle.
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=11, index=1, end_of_transfer=False, payload=hedgehog[50:]),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),  # OLD TID
            mk_frame(transfer_id=0, index=0, end_of_transfer=False, payload=hedgehog[50:]),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),  # LAST FRAME
            mk_frame(transfer_id=11, index=2, end_of_transfer=True, payload=TransferCRC.new(hedgehog).value_as_bytes),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),  # DUPLICATE OF INDEX 1
            mk_frame(transfer_id=11, index=1, end_of_transfer=False, payload=hedgehog[50:]),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),  # OLD TID
            mk_frame(transfer_id=10, index=1, end_of_transfer=False, payload=hedgehog[50:]),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),  # MALFORMED FRAME (no payload), ignored
            mk_frame(transfer_id=9999999999, index=0, end_of_transfer=False, payload=b""),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),  # FIRST FRAME
            mk_frame(transfer_id=11, index=0, end_of_transfer=False, payload=hedgehog[:50]),
        )
        == mk_transfer(timestamp=mk_ts(1000.0), transfer_id=11, fragmented_payload=[hedgehog[:50], hedgehog[50:]])
    )

    # Valid multi-frame transfer with payload size above the limit.
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=102, index=0, end_of_transfer=False, payload=hedgehog),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=102, index=1, end_of_transfer=False, payload=hedgehog),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=102, index=2, end_of_transfer=False, payload=hedgehog),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(
                transfer_id=102,
                index=3,
                end_of_transfer=True,
                payload=hedgehog + TransferCRC.new(hedgehog * 4).value_as_bytes,
            ),
        )
        == mk_transfer(timestamp=mk_ts(1000.0), transfer_id=102, fragmented_payload=[hedgehog] * 2)
    )

    # Same as above, but the frames are reordered.
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=103, index=2, end_of_transfer=False, payload=horse),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(
                transfer_id=103,
                index=3,
                end_of_transfer=True,
                payload=horse + TransferCRC.new(horse * 4).value_as_bytes,
            ),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=103, index=1, end_of_transfer=False, payload=horse),
        )
        is None
    )
    assert (
        push(
            mk_ts(1000.0),
            mk_frame(transfer_id=103, index=0, end_of_transfer=False, payload=horse),
        )
        == mk_transfer(timestamp=mk_ts(1000.0), transfer_id=103, fragmented_payload=[horse] * 2)
    )

    # Transfer-ID timeout. No error registered.
    assert (
        push(
            mk_ts(2000.0),
            mk_frame(transfer_id=0, index=0, end_of_transfer=True, payload=hedgehog),
        )
        == mk_transfer(timestamp=mk_ts(2000.0), transfer_id=0, fragmented_payload=[hedgehog])
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 0,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 0,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 0,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 0,
    }

    # Start a transfer, then start a new one with higher TID.
    assert (
        push(
            mk_ts(3000.0),  # Middle of a new transfer.
            mk_frame(transfer_id=2, index=1, end_of_transfer=False, payload=hedgehog),
        )
        is None
    )
    assert (
        push(
            mk_ts(3000.0),  # Another transfer! The old one is discarded.
            mk_frame(transfer_id=3, index=1, end_of_transfer=False, payload=horse[50:]),
        )
        is None
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 1,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 0,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 0,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 0,
    }
    assert (
        push(
            mk_ts(3000.0),
            mk_frame(transfer_id=3, index=2, end_of_transfer=True, payload=TransferCRC.new(horse).value_as_bytes),
        )
        is None
    )
    assert (
        push(
            mk_ts(3000.0),
            mk_frame(transfer_id=3, index=0, end_of_transfer=False, payload=horse[:50]),
        )
        == mk_transfer(timestamp=mk_ts(3000.0), transfer_id=3, fragmented_payload=[horse[:50], horse[50:]])
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 1,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 0,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 0,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 0,
    }

    # Start a transfer, then start a new one with lower TID when a TID timeout is reached.
    assert (
        push(
            mk_ts(3000.0),  # Middle of a new transfer.
            mk_frame(transfer_id=10, index=1, end_of_transfer=False, payload=hedgehog),
        )
        is None
    )
    assert (
        push(
            mk_ts(4000.0),  # Another transfer! The old one is discarded.
            mk_frame(transfer_id=3, index=1, end_of_transfer=False, payload=horse[50:]),
        )
        is None
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 2,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 0,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 0,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 0,
    }
    assert (
        push(
            mk_ts(4000.0),
            mk_frame(transfer_id=3, index=2, end_of_transfer=True, payload=TransferCRC.new(horse).value_as_bytes),
        )
        is None
    )
    assert (
        push(
            mk_ts(4000.0),
            mk_frame(transfer_id=3, index=0, end_of_transfer=False, payload=horse[:50]),
        )
        == mk_transfer(timestamp=mk_ts(4000.0), transfer_id=3, fragmented_payload=[horse[:50], horse[50:]])
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 2,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 0,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 0,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 0,
    }

    # Multi-frame transfer with bad CRC.
    assert (
        push(
            mk_ts(5000.0),
            mk_frame(transfer_id=10, index=1, end_of_transfer=False, payload=hedgehog[50:]),
        )
        is None
    )
    assert (
        push(
            mk_ts(5000.0),  # LAST FRAME
            mk_frame(
                transfer_id=10, index=2, end_of_transfer=True, payload=TransferCRC.new(hedgehog).value_as_bytes[::-1]
            ),  # Bad CRC here.
        )
        is None
    )
    assert (
        push(
            mk_ts(5000.0),  # FIRST FRAME
            mk_frame(transfer_id=10, index=0, end_of_transfer=False, payload=hedgehog[:50]),
        )
        is None
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 2,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 1,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 0,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 0,
    }

    # Frame past end of transfer.
    assert (
        push(
            mk_ts(5000.0),
            mk_frame(transfer_id=11, index=1, end_of_transfer=False, payload=hedgehog[50:]),
        )
        is None
    )
    assert (
        push(
            mk_ts(5000.0),  # PAST THE END OF TRANSFER
            mk_frame(transfer_id=11, index=3, end_of_transfer=False, payload=horse),
        )
        is None
    )
    assert (
        push(
            mk_ts(5000.0),  # LAST FRAME
            mk_frame(
                transfer_id=11, index=2, end_of_transfer=True, payload=TransferCRC.new(hedgehog + horse).value_as_bytes
            ),
        )
        is None
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 2,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 1,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 1,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 0,
    }

    # Inconsistent end-of-transfer flag.
    assert (
        push(
            mk_ts(5000.0),
            mk_frame(transfer_id=12, index=0, end_of_transfer=False, payload=hedgehog[:50]),
        )
        is None
    )
    assert (
        push(
            mk_ts(5000.0),  # LAST FRAME A
            mk_frame(
                transfer_id=12, index=2, end_of_transfer=True, payload=TransferCRC.new(hedgehog + horse).value_as_bytes
            ),
        )
        is None
    )
    assert (
        push(
            mk_ts(5000.0),  # LAST FRAME B
            mk_frame(transfer_id=12, index=3, end_of_transfer=True, payload=horse),
        )
        is None
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 2,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 1,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 1,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 1,
    }

    # Valid single-frame transfer with no payload.
    assert (
        push(
            mk_ts(6000.0),
            mk_frame(transfer_id=0, index=0, end_of_transfer=True, payload=b""),
        )
        == mk_transfer(timestamp=mk_ts(6000.0), transfer_id=0, fragmented_payload=[b""])
    )
    assert error_counters == {
        ta.Error.MULTIFRAME_MISSING_FRAMES: 2,
        ta.Error.MULTIFRAME_EMPTY_FRAME: 1,
        ta.Error.MULTIFRAME_INTEGRITY_ERROR: 1,
        ta.Error.MULTIFRAME_EOT_MISPLACED: 1,
        ta.Error.MULTIFRAME_EOT_INCONSISTENT: 1,
    }


def _unittest_transfer_reassembler_anonymous() -> None:
    ts = Timestamp.now()
    prio = Priority.LOW
    assert TransferReassembler.construct_anonymous_transfer(
        ts,
        Frame(priority=prio, transfer_id=123456, index=0, end_of_transfer=True, payload=memoryview(b"abcdef")),
    ) == TransferFrom(
        timestamp=ts, priority=prio, transfer_id=123456, fragmented_payload=[memoryview(b"abcdef")], source_node_id=None
    )

    assert (
        TransferReassembler.construct_anonymous_transfer(
            ts,
            Frame(priority=prio, transfer_id=123456, index=1, end_of_transfer=True, payload=memoryview(b"abcdef")),
        )
        is None
    )

    assert (
        TransferReassembler.construct_anonymous_transfer(
            ts,
            Frame(priority=prio, transfer_id=123456, index=0, end_of_transfer=False, payload=memoryview(b"abcdef")),
        )
        is None
    )


def _unittest_validate_and_finalize_transfer() -> None:
    ts = Timestamp.now()
    prio = Priority.FAST
    tid = 888888888
    src_nid = 1234

    def mk_transfer(fp: typing.Sequence[bytes]) -> TransferFrom:
        return TransferFrom(
            timestamp=ts,
            priority=prio,
            transfer_id=tid,
            fragmented_payload=list(map(memoryview, fp)),
            source_node_id=src_nid,
        )

    def call(fp: typing.Sequence[bytes]) -> typing.Optional[TransferFrom]:
        return _validate_and_finalize_transfer(
            timestamp=ts,
            priority=prio,
            transfer_id=tid,
            frame_payloads=list(map(memoryview, fp)),
            source_node_id=src_nid,
        )

    assert call([b""]) == mk_transfer([b""])
    assert call([b"hello world"]) == mk_transfer([b"hello world"])
    assert call(
        [b"hello world", b"0123456789", TransferCRC.new(b"hello world", b"0123456789").value_as_bytes]
    ) == mk_transfer([b"hello world", b"0123456789"])
    assert call([b"hello world", b"0123456789"]) is None  # no CRC


def _unittest_drop_crc() -> None:
    mv = memoryview
    assert _drop_crc([mv(b"0123456789")]) == [mv(b"012345")]
    assert _drop_crc([mv(b"0123456789"), mv(b"abcde")]) == [mv(b"0123456789"), mv(b"a")]
    assert _drop_crc([mv(b"0123456789"), mv(b"abcd")]) == [mv(b"0123456789")]
    assert _drop_crc([mv(b"0123456789"), mv(b"abc")]) == [mv(b"012345678")]
    assert _drop_crc([mv(b"0123456789"), mv(b"ab")]) == [mv(b"01234567")]
    assert _drop_crc([mv(b"0123456789"), mv(b"a")]) == [mv(b"0123456")]
    assert _drop_crc([mv(b"0123456789"), mv(b"")]) == [mv(b"012345")]
    assert _drop_crc([mv(b"0123456789"), mv(b""), mv(b"a"), mv(b"b")]) == [mv(b"01234567")]
    assert _drop_crc([mv(b"01"), mv(b""), mv(b"a"), mv(b"b")]) == []
    assert _drop_crc([mv(b"0"), mv(b""), mv(b"a"), mv(b"b")]) == []
    assert _drop_crc([mv(b"")]) == []
    assert _drop_crc([]) == []



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

r"""
UAVCAN/UDP transport overview
+++++++++++++++++++++++++++++

The UAVCAN/UDP transport is experimental and is not yet part of the UAVCAN specification.
Future revisions may break wire compatibility until the transport is formally specified.
Context: https://forum.uavcan.org/t/alternative-transport-protocols-in-uavcan/324/45?u=pavel.kirienko.

The UAVCAN/UDP transport is essentially a trivial stateless UDP blaster based on IP multicasting.
This transport is intended for low-latency, high-throughput switched Ethernet networks with complex topologies.
In the spirit of UAVCAN, it is designed to be simple and robust;
much of the data handling work is offloaded to the standard underlying UDP/IP stack.
Both IPv4 and IPv6 are supported by this design,
although it is expected that the advantages of IPv6 over IPv4 are less relevant in an intravehicular setting.

The concept of anonymous transfer is not defined for UAVCAN/UDP;
in this transport, in order to be able to emit a transfer, the node shall have a valid node-ID value.
This means that an anonymous UAVCAN/UDP node can only listen to network traffic
(i.e., can subscribe to subjects) but cannot transmit anything.
If address auto-configuration is desired, lower-level solutions should be used, such as DHCP.

This transport module contains no media sublayers because the media abstraction
is handled directly by the standard UDP/IP stack of the underlying operating system.

Per the UAVCAN transport model provided in the UAVCAN specification, the following transfer categories are supported:

+--------------------+--------------------------+---------------------------+
| Supported transfers| Unicast                  | Broadcast                 |
+====================+==========================+===========================+
|**Message**         | No                       | Yes                       |
+--------------------+--------------------------+---------------------------+
|**Service**         | Yes                      | Banned by Specification   |
+--------------------+--------------------------+---------------------------+


Protocol definition
+++++++++++++++++++

The entirety of the session specifier (:class:`pyuavcan.transport.SessionSpecifier`)
is reified through the standard UDP/IP stack without any special extensions.
The transfer-ID, transfer priority, and the multi-frame transfer reassembly metadata are allocated in the
UAVCAN-specific UDP datagram header.

+---------------------------------------+---------------------------------------------------------------------------+
| Parameter                             | Manifested in                                                             |
+=======================================+===========================================================================+
| Transfer priority                     |                                                                           |
+---------------------------------------+ UDP datagram payload (frame header)                                       |
| Transfer-ID                           |                                                                           |
+-------------------+-------------------+---------------------------------------------------------------------------+
|                   | Route specifier   | 16 least significant bits of the IP address                               |
| Session specifier +-------------------+---------------------------------------------------------------------------+
|                   | Data specifier    | For message transfers: 16 least significant bits of the                   |
|                   |                   | multicast group address.                                                  |
|                   |                   | For service transfers: UDP destination port number.                       |
+-------------------+-------------------+---------------------------------------------------------------------------+

There are two data types that model UAVCAN/UDP protocol data: :class:`UDPFrame` and :class:`RawPacket`.
The latter is never used during normal operation but only during on-line capture sessions
for reporting captured packets (see :class:`UDPCaptured`).


IP address mapping
~~~~~~~~~~~~~~~~~~

The IPv4 address of a node is structured as follows::

   xxxxxxxx.xddddddd.nnnnnnnn.nnnnnnnn
   \________/\_____/ \_______________/
    (9 bits) (7 bits)     (16 bits)
     prefix  subnet-ID     node-ID

Incoming traffic from IP addresses whose 16 most significant bits are different is rejected;
this behavior enables co-existence of multiple independent UAVCAN/UDP networks along with other UDP protocols
on the same network.

The *subnet-ID* is used to differentiate independent UAVCAN/UDP transport networks sharing the same IP network
(e.g., multiple UAVCAN/UDP networks running on localhost or on some physical network).
This is similar to the domain identifier in DDS.
This value is not used anywhere else in the protocol other than in the construction of the multicast group address,
as will be shown below.


Message transfers
~~~~~~~~~~~~~~~~~

Message transfers are executed as IP multicast transfers.
The IPv4 multicast group address is computed statically as follows::

       fixed         reserved
      (9 bits)       (3 bits)
      ________          _
     /        \        / \
     11101111.0ddddddd.000sssss.ssssssss
     \__/      \_____/    \____________/
   (4 bits)    (7 bits)      (13 bits)
     IPv4      subnet-ID     subject-ID
   multicast   \_______________________/
    prefix             (23 bits)
               collision-free multicast
                  addressing limit of
                 Ethernet MAC for IPv4

From the most significant bit to the least significant bit, the IPv4 multicast group address components are as follows:

- IPv4 multicast prefix is defined by RFC 1112.

- The following 5 bits are set to 0b11110 by this Specification. The motivation is as follows:

  - Setting the four least significant bits of the most significant byte to 0b1111 moves the address range
    into the administratively-scoped range (239.0.0.0/8, RFC 2365),
    which ensures that there may be no conflicts with well-known multicast groups.

  - Setting the most significant bit of the second octet to zero ensures that there may be no conflict
    with reserved sub-ranges within the administratively-scoped range.
    The resulting range 239.0.0.0/9 is entirely ad-hoc defined.

  - Fixing the 5+4=9 most significant bits of the multicast group address ensures that the variability
    is confined to the 23 least significant bits of the address only,
    which is desirable because the IPv4 Ethernet MAC layer does not differentiate beyond the
    23 least significant bits of the multicast group address
    (i.e., addresses that differ only in the 9 MSb collide at the MAC layer,
    which is unacceptable in a real-time system; see RFC 1112 section 6.4).
    Without this limitation, an engineer deploying a network might inadvertently create a configuration that
    causes MAC-layer collisions which may be difficult to detect.

- The following 7 bits (the least significant bits of the second octet) are used to differentiate
  independent UAVCAN/UDP networks sharing the same physical IP network.
  Since the 9 most significant bits of the node IP address are not represented in the multicast group address,
  nodes whose IP addresses differ only by the 9 MSb are not distinguished by UAVCAN/UDP.
  This limitation does not appear to be significant, though, because such configurations are easy to avoid.
  It follows that there may be up to 128 independent UAVCAN/UDP networks sharing the same IP subnet.

- The following 16 bits define the data specifier:

  - 3 bits reserved for future use.

  - 13 bits represent the subject-ID as-is.

Per RFC 1112, the default TTL is 1, which is unacceptable.
Therefore, publishers should use the TTL value of 16 by default,
which is chosen as a sensible default suitable for any intravehicular network.

Per RFC 1112, in order to emit a multicast packet, a limited level-1 implementation without the full support of
IGMP and multicast-specific packet handling policies is sufficient.

Due to the dependency on the dynamic IGMP configuration,
a newly configured subscriber may not immediately receive data from the subject --
a brief *subscription initialization latency* may occur (typically it is well under one second).
This is because the underlying IP stack needs to inform the network switch/router about its interest in a particular
multicast group by sending an IGMP membership report first.
A high-integrity application may choose to rely on a static switch configuration,
in which case no initialization delay will take place.

Example::

    Node IP address:    01111111 00000010 00000000 00001000
                             127        2        0        8

    Subject-ID:                              00010 00101010

    Multicast group:    11101111 00000010 00000010 00101010
                             239        2        2       42

Example::

    Node IP address:    11000000 10101000 00000000 00000001
                             192      168        0        1

    Subject-ID:                              00010 00101010

    Multicast group:    11101111 00101000 00000010 00101010
                             239       40        2       42


Service transfers
~~~~~~~~~~~~~~~~~

Service transfers are executed as regular IP unicast transfers.

The service data specifier (:class:`pyuavcan.transport.ServiceDataSpecifier`)
is manifested on the wire as the destination UDP port number;
the mapping function is implemented in :func:`udp_port_from_data_specifier`.
The source port number can be arbitrary (ephemeral), its value is ignored.

UAVCAN uses a wide range of UDP ports.
UDP/IP stacks that comply with the IANA ephemeral port range recommendations are expected to be
compatible with this; otherwise, there may be port assignment conflicts.
This, however, is not a problem for any major modern OS.


Datagram header format
~~~~~~~~~~~~~~~~~~~~~~

Every UAVCAN/UDP frame contains the following header before the payload,
encoded in the little-endian byte order, expressed here in the DSDL notation::

    uint8 version           # =0 in this revision; ignore frame otherwise.
    uint8 priority          # Like in CAN: 0 -- highest priority, 7 -- lowest priority.
    void16                  # Set to zero when transmitting, ignore when receiving.
    uint32 frame_index_eot  # MSB is set if the current frame is the last frame of the transfer.
    uint64 transfer_id      # The transfer-ID never overflows.
    void64                  # This space may be used later for runtime type identification.

The 31 least significant bits of the field ``frame_index_eot`` contain the frame index within the current transfer;
the most significant bit (31st) is set if the current frame is the last frame of the transfer.
Also see the documentation for :class:`UDPFrame`.

Multi-frame transfers contain four bytes of CRC32-C (Castagnoli) at the end computed over the entire transfer payload.
For more info on multi-frame transfers, please see
:class:`pyuavcan.transport.commons.high_overhead_transport.TransferReassembler`.


Unreliable networks and temporal redundancy
+++++++++++++++++++++++++++++++++++++++++++

For unreliable networks, deterministic data loss mitigation is supported.
This measure is only available for service transfers, not for message transfers due to their different semantics.
If the probability of a frame loss exceeds the desired reliability threshold,
the transport can be configured to repeat every outgoing service transfer a specified number of times,
on the assumption that the probability of losing any given frame is uncorrelated (or weakly correlated)
with that of its neighbors.

Assuming that the probability of transfer loss ``P`` is time-invariant,
the influence of the multiplier ``M`` can be approximated as ``P' = P^M``.
For example, given a network that successfully delivers 99% of transfers,
and the probabilities of adjacent transfer loss are uncorrelated,
the multiplication factor of 2 can increase the link reliability up to ``100% - (100% - 99%)^2 = 99.99%``.

The duplicates are emitted immediately following the original transfer.
For example, suppose that a service transfer contains three frames, F0 to F2,
and the service transfer multiplication factor is two,
then the resulting frame sequence would be as follows::

    F0      F1      F2      F0      F1      F2
    \_______________/       \_______________/
       main copy             redundant copy
     (TX timestamp)      (never TX-timestamped)

    ------------------ time ------------------>

As shown on the diagram, if the transmission timestamping is requested, only the first copy is timestamped.
Further, any errors occurring during the transmission of redundant copies
may be silently ignored by the stack, provided that the main copy is transmitted successfully.

The resulting behavior in the provided example is that the transport network may
lose up to three unique frames without affecting the application.
In the following example, the frames F0 and F2 of the main copy are lost, but the transfer survives::

    F0 F1 F2 F0 F1 F2
    |  |  |  |  |  |
    x  |  x  |  |  \_____ F2 __________________________
       |     |  \________ F1 (redundant, discarded) x  \
       |     \___________ F0 ________________________  |
       \_________________ F1 ______________________  \ |
                                                   \ | |
    ----- time ----->                              v v v
                                                reassembled
                                                multi-frame
                                                 transfer

Removal of duplicate transfers at the opposite end of the link is natively guaranteed by the UAVCAN protocol;
no special activities are needed there (read the UAVCAN Specification for background).

For time-deterministic (real-time) networks this strategy is preferred over the conventional
confirmation-retry approach (e.g., the TCP model) because it results in more predictable
network load, lower worst-case latency, and is stateless (participants do not make assumptions
about the state of other agents involved in data exchange).


Implementation-specific details
+++++++++++++++++++++++++++++++

Applications relying on this particular transport implementation will be unable to detect a node-ID conflict on
the bus because the implementation discards all traffic originating from its own IP address.
This is a very environment-specific edge case resulting from certain peculiarities of the Berkeley socket API.
Other implementations of UAVCAN/UDP (particularly those for embedded systems) may not have this limitation.


Usage
+++++

..  doctest::
    :hide:

    >>> import tests
    >>> tests.asyncio_allow_event_loop_access_from_top_level()
    >>> from tests import doctest_await

Create two transport instances -- one with a node-ID, one anonymous:

>>> import asyncio
>>> import pyuavcan
>>> import pyuavcan.transport.udp
>>> tr_0 = pyuavcan.transport.udp.UDPTransport('127.9.1.42')
>>> tr_0.local_node_id                                             # Derived from the IP address: (1 << 8) + 42 = 298.
298
>>> tr_1 = pyuavcan.transport.udp.UDPTransport('127.9.15.254', local_node_id=None)  # Anonymous is only for listening.
>>> tr_1.local_node_id is None
True

Create an output and an input session:

>>> pm = pyuavcan.transport.PayloadMetadata(1024)
>>> ds = pyuavcan.transport.MessageDataSpecifier(111)
>>> pub = tr_0.get_output_session(pyuavcan.transport.OutputSessionSpecifier(ds, None), pm)
>>> pub.socket.getpeername()   # UDP port is fixed, and the multicast group address is computed as shown above.
('239.9.0.111', 16383)
>>> sub = tr_1.get_input_session(pyuavcan.transport.InputSessionSpecifier(ds, None), pm)

Send a transfer from one instance to the other:

>>> doctest_await(pub.send(pyuavcan.transport.Transfer(pyuavcan.transport.Timestamp.now(),
...                                                    pyuavcan.transport.Priority.LOW,
...                                                    1111,
...                                                    fragmented_payload=[]),
...                        asyncio.get_event_loop().time() + 1.0))
True
>>> doctest_await(sub.receive(asyncio.get_event_loop().time() + 1.0))
TransferFrom(..., transfer_id=1111, ...)
>>> tr_0.close()
>>> tr_1.close()


Tooling
+++++++

Run UAVCAN networks on the local loopback interface (``127.x.y.z/8``) or create virtual interfaces for testing.

Use Wireshark for monitoring and inspection.

Use netcat for trivial monitoring; e.g., listen to a UDP port like this: ``nc -ul 48469``.

List all open UDP ports on the local machine: ``netstat -vpaun`` (GNU/Linux).


Inheritance diagram
+++++++++++++++++++

.. inheritance-diagram:: pyuavcan.transport.udp._udp
                         pyuavcan.transport.udp._frame
                         pyuavcan.transport.udp._session._input
                         pyuavcan.transport.udp._session._output
                         pyuavcan.transport.udp._socket_reader
                         pyuavcan.transport.udp._tracer
   :parts: 1
"""

from ._udp import UDPTransport as UDPTransport
from ._udp import UDPTransportStatistics as UDPTransportStatistics

from ._session import UDPInputSession as UDPInputSession
from ._session import PromiscuousUDPInputSession as PromiscuousUDPInputSession
from ._session import SelectiveUDPInputSession as SelectiveUDPInputSession

from ._session import UDPInputSessionStatistics as UDPInputSessionStatistics
from ._session import PromiscuousUDPInputSessionStatistics as PromiscuousUDPInputSessionStatistics
from ._session import SelectiveUDPInputSessionStatistics as SelectiveUDPInputSessionStatistics

from ._session import UDPOutputSession as UDPOutputSession
from ._session import UDPFeedback as UDPFeedback

from ._frame import UDPFrame as UDPFrame

from ._ip import IP_ADDRESS_NODE_ID_MASK as IP_ADDRESS_NODE_ID_MASK
from ._ip import SUBJECT_PORT as SUBJECT_PORT
from ._ip import node_id_to_unicast_ip as node_id_to_unicast_ip
from ._ip import unicast_ip_to_node_id as unicast_ip_to_node_id
from ._ip import message_data_specifier_to_multicast_group as message_data_specifier_to_multicast_group
from ._ip import multicast_group_to_message_data_specifier as multicast_group_to_message_data_specifier
from ._ip import service_data_specifier_to_udp_port as service_data_specifier_to_udp_port
from ._ip import udp_port_to_service_data_specifier as udp_port_to_service_data_specifier
from ._ip import LinkLayerPacket as LinkLayerPacket

from ._tracer import IPPacket as IPPacket
from ._tracer import IPv4Packet as IPv4Packet
from ._tracer import IPv6Packet as IPv6Packet
from ._tracer import UDPIPPacket as UDPIPPacket
from ._tracer import UDPCapture as UDPCapture
from ._tracer import UDPTracer as UDPTracer
from ._tracer import UDPErrorTrace as UDPErrorTrace



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import enum
from typing import Sequence
import pyuavcan
from pyuavcan.transport import Timestamp, TransferFrom
from .._frame import UAVCANFrame, compute_transfer_id_forward_distance, TRANSFER_CRC_LENGTH_BYTES, TRANSFER_ID_MODULO


class TransferReassemblyErrorID(enum.Enum):
    """
    Transfer reassembly error codes. Used in the extended error statistics.
    See the UAVCAN specification for background info.
    We have ``ID`` in the name to make clear that this is not an exception type.
    """

    MISSED_START_OF_TRANSFER = enum.auto()
    UNEXPECTED_TOGGLE_BIT = enum.auto()
    UNEXPECTED_TRANSFER_ID = enum.auto()
    TRANSFER_CRC_MISMATCH = enum.auto()


class TransferReassembler:
    def __init__(self, source_node_id: int, extent_bytes: int):
        self._source_node_id = int(source_node_id)
        self._timestamp = Timestamp(0, 0)
        self._transfer_id = 0
        self._toggle_bit = False
        self._max_payload_size_bytes_with_crc = int(extent_bytes) + TRANSFER_CRC_LENGTH_BYTES
        self._crc = pyuavcan.transport.commons.crc.CRC16CCITT()
        self._payload_truncated = False
        self._fragmented_payload: list[memoryview] = []

    def process_frame(
        self,
        timestamp: Timestamp,
        priority: pyuavcan.transport.Priority,
        frame: UAVCANFrame,
        transfer_id_timeout_ns: int,
    ) -> None | TransferReassemblyErrorID | TransferFrom:
        """
        Observe that occasionally newer frames may have lower timestamp values due to error variations in the time
        recovery algorithms, depending on the methods of timestamping. This class therefore does not check if the
        timestamp values are monotonically increasing. The timestamp of a transfer will be the lowest (earliest)
        timestamp value of its frames (ignoring frames with mismatching transfer ID or toggle bit).
        """
        # FIRST STAGE - DETECTION OF NEW TRANSFERS.
        # Decide if we need to begin a new transfer.
        tid_timed_out = (
            timestamp.monotonic_ns - self._timestamp.monotonic_ns > transfer_id_timeout_ns
            or self._timestamp.monotonic_ns == 0
        )

        not_previous_tid = compute_transfer_id_forward_distance(frame.transfer_id, self._transfer_id) > 1

        if tid_timed_out or (frame.start_of_transfer and not_previous_tid):
            self._transfer_id = frame.transfer_id
            self._toggle_bit = frame.toggle_bit
            if not frame.start_of_transfer:
                return TransferReassemblyErrorID.MISSED_START_OF_TRANSFER

        # SECOND STAGE - DROP UNEXPECTED FRAMES.
        # A properly functioning CAN bus may occasionally replicate frames (see the Specification for background).
        # We combat these issues by checking the transfer ID and the toggle bit.
        if frame.transfer_id != self._transfer_id:
            return TransferReassemblyErrorID.UNEXPECTED_TRANSFER_ID

        if frame.toggle_bit != self._toggle_bit:
            return TransferReassemblyErrorID.UNEXPECTED_TOGGLE_BIT

        # THIRD STAGE - PAYLOAD REASSEMBLY AND VERIFICATION.
        # Collect the data and check its correctness.
        if frame.start_of_transfer:
            self._crc = pyuavcan.transport.commons.crc.CRC16CCITT()
            self._payload_truncated = False
            self._fragmented_payload.clear()
            self._timestamp = timestamp  # Initialization from the first frame

        if self._timestamp.monotonic_ns > timestamp.monotonic_ns or self._timestamp.system_ns > timestamp.system_ns:
            # The timestamping algorithm may have corrected the time error since the first frame, accept lower value
            self._timestamp = Timestamp.combine_oldest(self._timestamp, timestamp)

        self._toggle_bit = not self._toggle_bit
        # Implicit truncation rule - discard the unexpected data at the end of the payload but compute the CRC anyway.
        self._crc.add(frame.padded_payload)
        if sum(map(len, self._fragmented_payload)) < self._max_payload_size_bytes_with_crc:
            self._fragmented_payload.append(frame.padded_payload)
        else:
            self._payload_truncated = True

        if frame.end_of_transfer:
            fragmented_payload = self._fragmented_payload.copy()
            self._prepare_for_next_transfer()
            self._fragmented_payload.clear()

            if frame.start_of_transfer:
                assert len(fragmented_payload) == 1  # Single-frame transfer, additional checks not needed
            else:
                # Multi-frame transfer, check and remove the trailing CRC.
                # We don't bother checking the CRC if we received fewer than 2 frames because that implies that there
                # was a TID wraparound mid-transfer.
                # This happens when the reassembler that has just been reset is fed with the last frame of another
                # transfer, whose TOGGLE and TRANSFER-ID happen to match the expectations of the reassembler:
                #   1. Wait for the reassembler to be reset. Let: expected transfer-ID = X, expected toggle bit = Y.
                #   2. Construct a frame with SOF=0, EOF=1, TID=X, TOGGLE=Y.
                #   3. Feed the frame into the reassembler.
                # See https://github.com/UAVCAN/pyuavcan/issues/198. There is a dedicated test covering this case.
                if len(fragmented_payload) < 2 or not self._crc.check_residue():
                    return TransferReassemblyErrorID.TRANSFER_CRC_MISMATCH

                # Cut off the CRC, unless it's already been removed by the implicit payload truncation rule.
                if not self._payload_truncated:
                    expected_length = sum(map(len, fragmented_payload)) - TRANSFER_CRC_LENGTH_BYTES
                    if len(fragmented_payload[-1]) > TRANSFER_CRC_LENGTH_BYTES:
                        fragmented_payload[-1] = fragmented_payload[-1][:-TRANSFER_CRC_LENGTH_BYTES]
                    else:
                        cutoff = TRANSFER_CRC_LENGTH_BYTES - len(fragmented_payload[-1])
                        assert cutoff >= 0
                        fragmented_payload = fragmented_payload[:-1]  # Drop the last fragment
                        if cutoff > 0:
                            fragmented_payload[-1] = fragmented_payload[-1][:-cutoff]  # Truncate the previous fragment
                    assert expected_length == sum(map(len, fragmented_payload))

            return TransferFrom(
                timestamp=self._timestamp,
                priority=priority,
                transfer_id=frame.transfer_id,
                fragmented_payload=fragmented_payload,
                source_node_id=self._source_node_id,
            )

        return None  # Expect more frames to come

    def _prepare_for_next_transfer(self) -> None:
        self._transfer_id = (self._transfer_id + 1) % TRANSFER_ID_MODULO
        self._toggle_bit = True


def _unittest_can_transfer_reassembler_manual() -> None:
    priority = pyuavcan.transport.Priority.IMMEDIATE
    source_node_id = 123
    transfer_id_timeout_ns = 900
    can_identifier = 0xBADC0FE

    err = TransferReassemblyErrorID

    def proc(monotonic_ns: int, frame: UAVCANFrame) -> None | TransferReassemblyErrorID | TransferFrom:
        away = rx.process_frame(
            timestamp=Timestamp(system_ns=0, monotonic_ns=monotonic_ns),
            priority=priority,
            frame=frame,
            transfer_id_timeout_ns=transfer_id_timeout_ns,
        )
        assert away is None or isinstance(away, (TransferReassemblyErrorID, TransferFrom))
        return away

    def frm(
        padded_payload: bytes | str,
        transfer_id: int,
        start_of_transfer: bool,
        end_of_transfer: bool,
        toggle_bit: bool,
    ) -> UAVCANFrame:
        return UAVCANFrame(
            identifier=can_identifier,
            padded_payload=memoryview(padded_payload if isinstance(padded_payload, bytes) else padded_payload.encode()),
            transfer_id=transfer_id,
            start_of_transfer=start_of_transfer,
            end_of_transfer=end_of_transfer,
            toggle_bit=toggle_bit,
        )

    def trn(
        monotonic_ns: int, transfer_id: int, fragmented_payload: Sequence[bytes | str | memoryview]
    ) -> TransferFrom:
        return TransferFrom(
            timestamp=Timestamp(system_ns=0, monotonic_ns=monotonic_ns),
            priority=priority,
            transfer_id=transfer_id,
            fragmented_payload=[
                memoryview(x if isinstance(x, (bytes, memoryview)) else x.encode()) for x in fragmented_payload
            ],
            source_node_id=source_node_id,
        )

    rx = TransferReassembler(source_node_id, 50)

    # Correct single-frame transfers.
    assert proc(1000, frm("Hello", 0, True, True, True)) == trn(1000, 0, ["Hello"])
    assert proc(1000, frm("Hello", 0, True, True, True)) == err.UNEXPECTED_TRANSFER_ID
    assert proc(1000, frm("Hello", 0, True, True, True)) == err.UNEXPECTED_TRANSFER_ID
    assert proc(2000, frm("Hello", 0, True, True, True)) == trn(2000, 0, ["Hello"])  # TID timeout

    # Correct multi-frame transfer.
    assert proc(2000, frm(b"\x00\x01\x02\x03\x04\x05\x06", 1, True, False, True)) is None
    assert proc(2001, frm(b"\x07\x08\x09\x0a\x0b\x0c\x0d", 1, False, False, False)) is None
    assert proc(2002, frm(b"\x0e\x0f\x10\x11\x12\x13\x14", 1, False, False, True)) is None
    assert proc(2003, frm(b"\x15\x16\x17\x18\x19\x1a\x1b", 1, False, False, False)) is None
    assert proc(2004, frm(b"\x1c\x1d\x35\x54", 1, False, True, True)) == trn(
        2000,
        1,
        [
            b"\x00\x01\x02\x03\x04\x05\x06",
            b"\x07\x08\x09\x0a\x0b\x0c\x0d",
            b"\x0e\x0f\x10\x11\x12\x13\x14",
            b"\x15\x16\x17\x18\x19\x1a\x1b",
            b"\x1c\x1d",
        ],
    )

    # Correct transfer with the old transfer ID will be ignored.
    assert proc(2010, frm(b"\x00\x01\x02\x03\x04\x05\x06", 1, True, False, True)) == err.UNEXPECTED_TRANSFER_ID
    assert proc(2011, frm(b"\x07\x08\x09\x0a\x0b\x0c\x0d", 1, False, False, False)) == err.UNEXPECTED_TRANSFER_ID
    assert proc(2012, frm(b"\x0e\x0f\x10\x11\x12\x13\x14", 1, False, False, True)) == err.UNEXPECTED_TRANSFER_ID
    assert proc(2013, frm(b"\x15\x16\x17\x18\x19\x1a\x1b", 1, False, False, False)) == err.UNEXPECTED_TRANSFER_ID
    assert proc(2014, frm(b"\x1c\x1d\x35\x54", 1, False, True, True)) == err.UNEXPECTED_TRANSFER_ID

    # Correct reassembly where the CRC spills over into the next frame.
    assert (
        proc(2100, frm(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e", 9, True, False, True)) is None
    )
    assert (
        proc(2101, frm(b"\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\xc4", 9, False, False, False)) is None
    )
    assert proc(2102, frm(b"\x6f", 9, False, True, True)) == trn(
        2100,
        9,
        [
            b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e",
            b"\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c",  # Third fragment is gone - used to contain CRC
        ],
    )

    # Transfer ID rolled back but should be accepted anyway; CRC is invalid
    assert (
        proc(2200, frm(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e", 8, True, False, True)) is None
    )
    assert (
        proc(2201, frm(b"\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\xc4", 8, False, False, False)) is None
    )
    assert proc(2202, frm(b"\x00", 8, False, True, True)) == err.TRANSFER_CRC_MISMATCH

    # Transfer ID timeout and the new frame is not a start of new transfer --> missed start error
    assert proc(4000, frm(b"123456", 8, False, False, True)) == err.MISSED_START_OF_TRANSFER

    # New transfer; same TID is accepted anyway due to the timeout condition; repeated frames (bad toggles)
    assert proc(4000, frm(b"\x00\x01\x02\x03\x04\x05\x06", 8, True, False, True)) is None
    assert proc(4010, frm(b"123456", 8, True, False, True)) == err.UNEXPECTED_TOGGLE_BIT
    assert proc(3500, frm(b"\x07\x08\x09\x0a\x0b\x0c\x0d", 8, False, False, False)) is None  # Timestamp update!
    assert proc(3000, frm(b"", 8, False, False, False)) == err.UNEXPECTED_TOGGLE_BIT  # Timestamp ignored
    assert proc(4022, frm(b"\x0e\x0f\x10\x11\x12\x13\x14", 8, False, False, True)) is None
    assert proc(4002, frm(b"\x0e\x0f\x10\x11\x12\x13\x14", 8, False, False, True)) == err.UNEXPECTED_TOGGLE_BIT
    assert proc(4013, frm(b"\x15\x16\x17\x18\x19\x1a\x1b", 8, False, False, False)) is None
    assert proc(4003, frm(b"\x15\x16\x17\x18\x19\x1a\x1b" * 2, 8, False, False, False)) == err.UNEXPECTED_TOGGLE_BIT
    assert proc(4004, frm(b"\x1c\x1d\x35\x54", 8, False, True, True)) == trn(
        3500,
        8,
        [
            b"\x00\x01\x02\x03\x04\x05\x06",
            b"\x07\x08\x09\x0a\x0b\x0c\x0d",
            b"\x0e\x0f\x10\x11\x12\x13\x14",
            b"\x15\x16\x17\x18\x19\x1a\x1b",
            b"\x1c\x1d",
        ],
    )
    assert proc(4004, frm(b"\x1c\x1d\x35\x54", 8, False, True, True)) == err.UNEXPECTED_TRANSFER_ID  # Not toggle!

    # Transfer that is too large (above the configured limit) is implicitly truncated. Time goes back but it's fine.
    assert proc(1000, frm(b"0123456789abcdefghi", 0, True, False, True)) is None  # 19
    assert proc(1001, frm(b"0123456789abcdefghi", 0, False, False, False)) is None  # 38
    assert proc(1001, frm(b"0123456789abcdefghi", 0, False, False, True)) is None  # 57
    assert proc(1001, frm(b"0123456789abcdefghi", 0, False, False, False)) is None  # 76
    assert proc(1001, frm(b":B", 0, False, True, True)) == trn(
        1000,
        0,
        [
            b"0123456789abcdefghi",
            b"0123456789abcdefghi",
            b"0123456789abcdefghi",
            # Last two are truncated away.
        ],
    )

    # Transfer above the limit but accepted nevertheless because the overflow induced by the last frame is not checked.
    assert proc(1000, frm(b"0123456789abcdefghi", 31, True, False, True)) is None  # 19
    assert proc(1001, frm(b"0123456789abcdefghi", 31, False, False, False)) is None  # 38
    assert proc(1001, frm(b"0123456789abcdefghi\xa9\x72", 31, False, True, True)) == trn(
        1000,
        31,
        [
            b"0123456789abcdefghi",
            b"0123456789abcdefghi",
            b"0123456789abcdefghi",
        ],
    )


def _unittest_issue_198() -> None:
    source_node_id = 88
    transfer_id_timeout_ns = 900

    def mk_frame(
        padded_payload: bytes | str,
        transfer_id: int,
        start_of_transfer: bool,
        end_of_transfer: bool,
        toggle_bit: bool,
    ) -> UAVCANFrame:
        return UAVCANFrame(
            identifier=0xBADC0FE,
            padded_payload=memoryview(padded_payload if isinstance(padded_payload, bytes) else padded_payload.encode()),
            transfer_id=transfer_id,
            start_of_transfer=start_of_transfer,
            end_of_transfer=end_of_transfer,
            toggle_bit=toggle_bit,
        )

    rx = TransferReassembler(source_node_id, 50)

    # First, ensure that the reassembler is initialized, by feeding it a valid transfer at least once.
    assert rx.process_frame(
        timestamp=Timestamp(system_ns=0, monotonic_ns=1000),
        priority=pyuavcan.transport.Priority.SLOW,
        frame=mk_frame("123", 0, True, True, True),
        transfer_id_timeout_ns=transfer_id_timeout_ns,
    ) == TransferFrom(
        timestamp=Timestamp(system_ns=0, monotonic_ns=1000),
        priority=pyuavcan.transport.Priority.SLOW,
        transfer_id=0,
        fragmented_payload=[memoryview(x if isinstance(x, (bytes, memoryview)) else x.encode()) for x in ["123"]],
        source_node_id=source_node_id,
    )

    # Next, feed the last frame of another transfer whose TID/TOG match the expected state of the reassembler.
    # This should be recognized as a CRC error.
    assert (
        rx.process_frame(
            timestamp=Timestamp(system_ns=0, monotonic_ns=1000),
            priority=pyuavcan.transport.Priority.SLOW,
            frame=mk_frame("456", 1, False, True, True),
            transfer_id_timeout_ns=transfer_id_timeout_ns,
        )
        == TransferReassemblyErrorID.TRANSFER_CRC_MISMATCH
    )
